from typing import List, Optional, Set, Tuple

from jinja2 import Template
from markupsafe import Markup
from starlette.datastructures import FormData

from strictdoc.backend.sdoc.models.document import SDocDocument
from strictdoc.backend.sdoc.models.document_grammar import (
    DocumentGrammar,
)
from strictdoc.backend.sdoc.models.grammar_element import (
    GrammarElement,
    GrammarElementField,
    GrammarElementFieldString,
    GrammarElementFieldType,
    GrammarElementRelationChild,
    GrammarElementRelationFile,
    GrammarElementRelationParent,
    GrammarElementRelationType,
)
from strictdoc.backend.sdoc.models.model import RequirementFieldName
from strictdoc.core.project_config import ProjectConfig
from strictdoc.export.html.form_objects.form_object import (
    RowWithCustomFieldFormObject,
    RowWithRelationFormObject,
    RowWithReservedFieldFormObject,
)
from strictdoc.export.html.html_templates import JinjaEnvironment
from strictdoc.helpers.auto_described import auto_described
from strictdoc.helpers.cast import assert_cast
from strictdoc.helpers.form_data import ParsedFormData, parse_form_data
from strictdoc.helpers.mid import MID
from strictdoc.helpers.string import is_uppercase_underscore_string
from strictdoc.server.error_object import ErrorObject
from strictdoc.server.helpers.turbo import render_turbo_stream


def is_reserved_field(field_name: str) -> bool:
    return field_name in (
        RequirementFieldName.UID,
        RequirementFieldName.TITLE,
        RequirementFieldName.STATEMENT,
        RequirementFieldName.RATIONALE,
        RequirementFieldName.COMMENT,
    )


@auto_described
class GrammarFormField:
    def __init__(
        self,
        field_mid: str,
        field_name: str,
        field_human_title: Optional[str],
        field_required: bool,
        reserved: bool,
    ):
        self.field_mid: str = field_mid
        self.field_name: str = field_name
        self.field_human_title: Optional[str] = field_human_title
        self.field_required: bool = field_required
        self.reserved: bool = reserved

    @staticmethod
    def create_from_grammar_field(
        *, grammar_field: GrammarElementField
    ) -> "GrammarFormField":
        reserved = is_reserved_field(grammar_field.title)
        return GrammarFormField(
            field_mid=grammar_field.mid,
            field_name=grammar_field.title,
            field_human_title=grammar_field.human_title,
            field_required=grammar_field.required,
            reserved=reserved,
        )

    def get_input_field_name(self) -> str:
        return f"document_grammar_field[{self.field_mid}][field_name]"

    def get_input_field_human_title(self) -> str:
        return f"document_grammar_field[{self.field_mid}][field_human_title]"

    def get_input_field_required(self) -> str:
        return f"document_grammar_field[{self.field_mid}][field_required]"

    def get_input_field_required_value(self) -> str:
        return "true" if self.field_required else "false"


@auto_described
class GrammarFormRelation:
    def __init__(
        self,
        relation_mid: str,
        relation_type: str,
        relation_role: Optional[str],
    ):
        self.relation_mid: str = relation_mid
        self.relation_type: str = relation_type
        self.relation_role: str = (
            relation_role if relation_role is not None else ""
        )

    def relation_type_input_name(self) -> str:
        return f"document_grammar_relation[{self.relation_mid}][type]"

    def relation_role_input_name(self) -> str:
        return f"document_grammar_relation[{self.relation_mid}][role]"


@auto_described
class GrammarElementFormObject(ErrorObject):
    def __init__(
        self,
        *,
        document_mid: str,
        element_mid: str,
        element_name: str,
        is_composite: Optional[bool],
        prefix: Optional[str],
        view_style: Optional[str],
        fields: List[GrammarFormField],
        relations: List[GrammarFormRelation],
        project_config: ProjectConfig,
        jinja_environment: JinjaEnvironment,
    ):
        assert isinstance(document_mid, str), document_mid
        super().__init__()
        self.document_mid = document_mid
        self.element_mid: str = element_mid
        self.element_name: str = element_name
        self.is_composite: Optional[bool] = is_composite
        self.prefix: Optional[str] = prefix
        self.view_style: Optional[str] = view_style
        self.fields: List[GrammarFormField] = fields
        self.relations: List[GrammarFormRelation] = relations
        self.project_config: ProjectConfig = project_config
        self.jinja_environment: JinjaEnvironment = jinja_environment

    @staticmethod
    def create_from_request(
        *,
        document: SDocDocument,
        request_form_data: FormData,
        project_config: ProjectConfig,
        jinja_environment: JinjaEnvironment,
    ) -> "GrammarElementFormObject":
        assert document.grammar is not None

        form_object_fields: List[GrammarFormField] = []
        form_object_relations: List[GrammarFormRelation] = []
        request_form_data_as_list = [
            (field_name, field_value)
            for field_name, field_value in request_form_data.multi_items()
        ]
        request_form_dict: ParsedFormData = assert_cast(
            parse_form_data(request_form_data_as_list), dict
        )

        element_mid = assert_cast(request_form_dict["element_mid"], str)

        element_is_composite = request_form_dict["is_composite"]
        element_prefix = request_form_dict["prefix"]
        element_view_style = request_form_dict["view_style"]

        #
        # Grammar fields.
        #
        document_grammar_fields = assert_cast(
            request_form_dict["document_grammar_field"], dict
        )
        for field_mid, field_dict in document_grammar_fields.items():
            assert isinstance(field_dict, dict), type(field_dict)

            field_name = assert_cast(field_dict["field_name"], str)
            field_human_title = field_dict.get("field_human_title")
            if field_human_title is not None:
                field_human_title = field_human_title.strip()
                if len(field_human_title) == 0:
                    field_human_title = None
            field_required_string = field_dict.get("field_required")
            field_required = field_required_string == "true"

            form_object_field = GrammarFormField(
                field_mid=field_mid,
                field_name=field_name,
                field_human_title=field_human_title,
                field_required=field_required,
                reserved=is_reserved_field(field_name),
            )
            form_object_fields.append(form_object_field)

        #
        # Grammar relations.
        #
        document_grammar_relations = assert_cast(
            request_form_dict.get("document_grammar_relation", {}), dict
        )
        for field_mid, field_dict in document_grammar_relations.items():
            assert isinstance(field_dict, dict), type(field_dict)

            field_type = assert_cast(field_dict["type"], str)
            field_role = assert_cast(field_dict["role"].strip(), str)

            form_object_relation = GrammarFormRelation(
                relation_mid=field_mid,
                relation_type=field_type,
                relation_role=field_role,
            )
            form_object_relations.append(form_object_relation)

        element: GrammarElement = document.grammar.get_element_by_mid(
            element_mid
        )

        form_object = GrammarElementFormObject(
            document_mid=document.reserved_mid,
            element_mid=element_mid,
            element_name=element.tag,
            is_composite=(
                True
                if element_is_composite == "true"
                else False
                if element_is_composite == "false"
                else None
            ),
            prefix=assert_cast(element_prefix, str),
            view_style=assert_cast(element_view_style, str),
            fields=form_object_fields,
            relations=form_object_relations,
            project_config=project_config,
            jinja_environment=jinja_environment,
        )
        return form_object

    @staticmethod
    def create_from_document(
        *,
        document: SDocDocument,
        element_mid: str,
        project_config: ProjectConfig,
        jinja_environment: JinjaEnvironment,
    ) -> "GrammarElementFormObject":
        assert isinstance(document, SDocDocument)
        assert isinstance(document.grammar, DocumentGrammar)

        grammar: DocumentGrammar = document.grammar

        element: GrammarElement = grammar.get_element_by_mid(element_mid)

        grammar_form_fields: List[GrammarFormField] = []
        for grammar_field in element.fields:
            grammar_form_field = GrammarFormField.create_from_grammar_field(
                grammar_field=grammar_field
            )
            grammar_form_fields.append(grammar_form_field)

        grammar_form_relations: List[GrammarFormRelation] = []
        for grammar_relation in element.relations:
            grammar_form_relation = GrammarFormRelation(
                relation_mid=grammar_relation.mid,
                relation_type=grammar_relation.relation_type,
                relation_role=grammar_relation.relation_role,
            )
            grammar_form_relations.append(grammar_form_relation)

        return GrammarElementFormObject(
            document_mid=document.reserved_mid,
            element_mid=element_mid,
            element_name=element.tag,
            is_composite=element.property_is_composite,
            prefix=element.property_prefix,
            view_style=element.property_view_style,
            fields=grammar_form_fields,
            relations=grammar_form_relations,
            project_config=project_config,
            jinja_environment=jinja_environment,
        )

    def validate(self) -> bool:
        fields_so_far: Set[str] = set()
        for field in self.fields:
            if len(field.field_name) == 0:
                self.add_error(
                    field.get_input_field_name(),
                    f"Grammar field {field.field_name} must not be empty.",
                )
                continue

            if not is_uppercase_underscore_string(field.field_name):
                self.add_error(
                    field.get_input_field_name(),
                    (
                        "Grammar field title shall consist of "
                        "uppercase letters, digits and single underscores."
                    ),
                )
                continue

            if field.field_name in fields_so_far:
                self.add_error(
                    field.get_input_field_name(),
                    f"Grammar field {field.field_name} is not unique.",
                )
            else:
                fields_so_far.add(field.field_name)

        if len(self.relations) == 0 and self.element_name not in (
            "TEXT",
            "SECTION",
        ):
            self.add_error(
                "Relations_Row",
                (
                    "Every grammar must include at least one relation. "
                    "A grammar lacking any relations is not considered a "
                    "realistic use case. To address this issue, you can create "
                    "a default 'Parent' relation with no assigned role."
                ),
            )
        else:
            general_relations_so_far: Set[str] = set()
            role_relations_so_far: Set[Tuple[str, str]] = set()
            for relation in self.relations:
                if (
                    relation.relation_role is None
                    or len(relation.relation_role) == 0
                ):
                    if relation.relation_type in general_relations_so_far:
                        self.add_error(
                            relation.relation_type_input_name(),
                            (
                                f"A duplicated general relation: {relation.relation_type}. "
                                "A relation is general when it does not have a role that specializes the relation. "
                                "A grammar must have zero or one general relations of a given type (e.g., Parent or Child)."
                            ),
                        )
                    else:
                        general_relations_so_far.add(relation.relation_type)
                else:
                    if (
                        relation.relation_type,
                        relation.relation_role,
                    ) in role_relations_so_far:
                        self.add_error(
                            relation.relation_type_input_name(),
                            (
                                f"A duplicated relation and role: "
                                f"{relation.relation_type} ({relation.relation_role}). "
                                "A grammar must have zero or one relations with a given type and a given role."
                            ),
                        )
                    else:
                        role_relations_so_far.add(
                            (relation.relation_type, relation.relation_role)
                        )

        return len(self.errors) == 0

    def convert_to_grammar_element(
        self, existing_grammar: DocumentGrammar
    ) -> GrammarElement:
        grammar_fields: List[GrammarElementFieldType] = []
        for field in self.fields:
            grammar_field = GrammarElementFieldString(
                parent=None,
                title=field.field_name,
                human_title=field.field_human_title,
                required="True" if field.field_required else "False",
            )
            grammar_fields.append(grammar_field)
        relation_fields: List[GrammarElementRelationType] = []

        for relation in self.relations:
            if relation.relation_type == "Parent":
                relation_fields.append(
                    GrammarElementRelationParent(
                        parent=None,
                        relation_type=relation.relation_type,
                        relation_role=relation.relation_role,
                    )
                )
            elif relation.relation_type == "Child":
                relation_fields.append(
                    GrammarElementRelationChild(
                        parent=None,
                        relation_type=relation.relation_type,
                        relation_role=relation.relation_role,
                    )
                )
            elif relation.relation_type == "File":
                relation_fields.append(
                    GrammarElementRelationFile(
                        parent=None,
                        relation_type=relation.relation_type,
                        relation_role=relation.relation_role,
                    )
                )
            else:
                raise NotImplementedError(relation)

        existing_element = existing_grammar.get_element_by_mid(self.element_mid)

        requirement_element = GrammarElement(
            parent=None,
            tag=existing_element.tag,
            property_is_composite=(
                "True"
                if self.is_composite == True
                else "False"
                if self.is_composite == False
                else ""
            ),
            property_prefix=self.prefix or "",
            property_view_style=self.view_style or "",
            fields=grammar_fields,
            relations=relation_fields,
        )
        return requirement_element

    def render(self) -> Markup:
        template: Template = self.jinja_environment.get_template(
            "components/grammar_form_element/index.jinja"
        )
        rendered_template = template.render(form_object=self)
        return render_turbo_stream(
            content=rendered_template, action="update", target="modal"
        )

    def render_after_validation(self) -> Markup:
        rendered_template = self.jinja_environment.render_template_as_markup(
            "components/grammar_form_element/index.jinja", form_object=self
        )
        return render_turbo_stream(
            content=rendered_template, action="update", target="modal"
        )

    def render_row_with_reserved_field(self, field: GrammarFormField) -> str:
        form_object = RowWithReservedFieldFormObject(
            field=field,
            errors=self.errors,
            jinja_environment=self.jinja_environment,
        )
        return form_object.render()

    def render_row_with_custom_field(self, field: GrammarFormField) -> str:
        assert isinstance(field, GrammarFormField)
        form_object = RowWithCustomFieldFormObject(
            field=field,
            errors=self.errors,
            jinja_environment=self.jinja_environment,
        )
        return form_object.render()

    def render_row_with_relation(self, relation: GrammarFormRelation) -> str:
        assert isinstance(relation, GrammarFormRelation)
        form_object = RowWithRelationFormObject(
            relation=relation,
            errors=self.errors,
            jinja_environment=self.jinja_environment,
        )
        return form_object.render()

    def render_row_with_new_field(self) -> str:
        field: GrammarFormField = GrammarFormField(
            field_mid=MID.create(),
            field_name="",
            field_human_title=None,
            field_required=False,
            reserved=False,
        )
        form_object = RowWithCustomFieldFormObject(
            field=field,
            errors=self.errors,
            jinja_environment=self.jinja_environment,
        )
        rendered_template: str = form_object.render()
        return render_turbo_stream(
            content=rendered_template,
            action="append",
            target="document__editable_grammar_fields",
        )

    def render_row_with_new_relation(self) -> str:
        relation = GrammarFormRelation(
            relation_mid=MID.create(),
            relation_type="Parent",
            relation_role="",
        )
        form_object = RowWithRelationFormObject(
            relation=relation,
            errors=self.errors,
            jinja_environment=self.jinja_environment,
        )
        rendered_template: str = form_object.render()
        return render_turbo_stream(
            content=rendered_template,
            action="append",
            target="document__editable_grammar_relations",
        )

    @staticmethod
    def render_close_form() -> str:
        return render_turbo_stream(
            content="",
            action="update",
            target="modal",
        )

    def get_input_field_is_composite(self) -> str:
        return "is_composite"

    def get_input_field_is_composite_value(self) -> str:
        if self.is_composite is not None:
            if self.is_composite == True:
                return "true"
            return "false"
        return ""

    def get_input_field_prefix(self) -> str:
        return "prefix"

    def get_input_field_prefix_value(self) -> str:
        return self.prefix or ""

    def get_input_field_view_style(self) -> str:
        return "view_style"

    def get_input_field_view_style_value(self) -> str:
        return self.view_style or ""



"""
@relation(SDOC-SRS-142, scope=file)
"""

from typing import Optional, Union

from robot.api.parsing import (
    Comment,
    Documentation,
    EmptyLine,
    ModelVisitor,
    Tags,
    TestCase,
    Token,
    get_model,
)

from strictdoc.backend.sdoc_source_code.constants import FunctionAttribute
from strictdoc.backend.sdoc_source_code.marker_parser import MarkerParser
from strictdoc.backend.sdoc_source_code.models.function import Function
from strictdoc.backend.sdoc_source_code.models.function_range_marker import (
    FunctionRangeMarker,
    RangeMarkerType,
)
from strictdoc.backend.sdoc_source_code.models.line_marker import LineMarker
from strictdoc.backend.sdoc_source_code.models.range_marker import (
    RangeMarker,
)
from strictdoc.backend.sdoc_source_code.models.source_file_info import (
    SourceFileTraceabilityInfo,
)
from strictdoc.backend.sdoc_source_code.parse_context import ParseContext
from strictdoc.backend.sdoc_source_code.processors.general_language_marker_processors import (
    function_range_marker_processor,
    line_marker_processor,
    range_marker_processor,
    source_file_traceability_info_processor,
)
from strictdoc.helpers.file_stats import SourceFileStats


class SdocRelationVisitor(ModelVisitor):  # type: ignore[misc]
    """
    Create functions from test cases in *.robot files and create markers.

    Note: ModelVisitor reuses ast.NodeVisitor from Python. We rely on following
    behavior.
    - It traverses depth first. This order is important to get matching marker
      pairs on the ParsersContext.marker_stack.
    - It doesn't recurse into subtrees if a custom visit_* method is defined.
      This is important to avoid duplicated matches.
    """

    def __init__(
        self,
        traceability_info: SourceFileTraceabilityInfo,
        parse_context: ParseContext,
    ):
        super().__init__()
        self.traceability_info = traceability_info
        self.parse_context = parse_context

    def visit_Comment(self, node: Comment) -> None:
        """
        Create non-function Marker from Comment outside TestCases.
        """
        self._visit_possibly_marked_node(node)

    def visit_Documentation(self, node: Documentation) -> None:
        """
        Create non-function Marker from Documentation outside TestCases.
        """
        self._visit_possibly_marked_node(node)

    def visit_Tags(self, node: Tags) -> None:
        """
        Create non-function Marker from Tags outside TestCases.
        """
        self._visit_possibly_marked_node(node)

    def visit_TestCase(self, node: TestCase) -> None:
        """
        Create function and non-function Marker from TestCases.
        """
        trailing_empty_lines = 0
        tc_markers = []
        for stmt in node.body:
            if isinstance(stmt, EmptyLine):
                # Trim trailing newlines from test case range.
                trailing_empty_lines += 1
            else:
                trailing_empty_lines = 0

            if isinstance(stmt, (Comment, Tags, Documentation)):
                for token in filter(self._token_filter, stmt.tokens):
                    source_node = MarkerParser.parse(
                        token.value,
                        token.lineno,
                        token.lineno,
                        token.lineno,
                        node.name,
                        token.col_offset,
                    )
                    tc_markers.extend(source_node.markers)

        function_markers = []
        for marker_ in tc_markers:
            if isinstance(marker_, FunctionRangeMarker):
                marker_.ng_range_line_begin = node.lineno
                marker_.ng_range_line_end = (
                    node.end_lineno - trailing_empty_lines
                )
                function_markers.append(marker_)
                function_range_marker_processor(marker_, self.parse_context)
            elif isinstance(marker_, RangeMarker):
                range_marker_processor(marker_, self.parse_context)
            elif isinstance(marker_, LineMarker):
                line_marker_processor(marker_, self.parse_context)

        self.traceability_info.markers.extend(function_markers)
        test_case = Function(
            parent=self.traceability_info,
            name=node.name,
            display_name=node.name,
            line_begin=node.lineno,
            line_end=node.end_lineno - trailing_empty_lines,
            child_functions=[],
            markers=function_markers,
            attributes={FunctionAttribute.DEFINITION},
        )
        self.traceability_info.functions.append(test_case)

    def _visit_possibly_marked_node(
        self, node: Union[Comment, Documentation, Tags]
    ) -> None:
        for token in filter(self._token_filter, node.tokens):
            source_node = MarkerParser.parse(
                token.value,
                node.lineno,
                node.lineno,
                node.lineno,
                None,
                token.col_offset,
            )
            for marker_ in source_node.markers:
                if (
                    isinstance(marker_, FunctionRangeMarker)
                    and marker_.scope is RangeMarkerType.FILE
                ):
                    # Outside Test Cases only accept scope=file function markers
                    marker_.ng_range_line_begin = 1
                    marker_.ng_range_line_end = (
                        self.parse_context.file_stats.lines_total
                    )
                    function_range_marker_processor(marker_, self.parse_context)
                    self.traceability_info.markers.append(marker_)
                elif isinstance(marker_, RangeMarker):
                    range_marker_processor(marker_, self.parse_context)
                elif isinstance(marker_, LineMarker):
                    line_marker_processor(marker_, self.parse_context)

    @staticmethod
    def _token_filter(token: Token) -> bool:
        if token.type in ("SEPARATOR", "EOL"):
            return False
        return True


class SourceFileTraceabilityReader_Robot:
    def read(
        self, input_buffer: str, file_path: Optional[str] = None
    ) -> SourceFileTraceabilityInfo:
        traceability_info = SourceFileTraceabilityInfo([])
        if len(input_buffer) == 0:
            return traceability_info
        file_stats = SourceFileStats.create(input_buffer)
        parse_context = ParseContext(file_path, file_stats)
        robotfw_model = get_model(input_buffer, data_only=False)
        visitor = SdocRelationVisitor(traceability_info, parse_context)
        visitor.visit(robotfw_model)
        source_file_traceability_info_processor(
            traceability_info, parse_context
        )
        return traceability_info

    def read_from_file(self, file_path: str) -> SourceFileTraceabilityInfo:
        with open(file_path) as file:
            sdoc_content = file.read()
            sdoc = self.read(sdoc_content, file_path=file_path)
            return sdoc



from jinja2 import Environment, FileSystemLoader, StrictUndefined

from strictdoc import environment


class RSTTemplates:
    jinja_environment = Environment(
        loader=FileSystemLoader(environment.get_path_to_rst_templates()),
        undefined=StrictUndefined,
    )
    # TODO: Check if this line is still needed (might be some older workaround).
    jinja_environment.globals.update(isinstance=isinstance)
    jinja_environment.trim_blocks = False
    jinja_environment.lstrip_blocks = False
    jinja_environment.keep_trailing_newline = True



# pylint: disable=wrong-import-position
# flake8: noqa: E402

# Needed to ensure that multiprocessing.freeze_support() is called
# in a frozen application (see main() below).
import multiprocessing
import os
import sys
from typing import Optional

strictdoc_root_path = os.path.abspath(
    os.path.join(os.path.dirname(__file__), "..", "..")
)
assert os.path.isdir(strictdoc_root_path)
sys.path.append(strictdoc_root_path)

from strictdoc import environment
from strictdoc.cli.cli_arg_parser import (
    CLIValidationError,
    DiffCommandConfig,
    DumpGrammarCommandConfig,
    ExportCommandConfig,
    ImportExcelCommandConfig,
    ImportReqIFCommandConfig,
    ManageAutoUIDCommandConfig,
    SDocArgsParser,
    create_sdoc_args_parser,
)
from strictdoc.commands.about_command import AboutCommand
from strictdoc.commands.diff_command import DiffCommand
from strictdoc.commands.dump_grammar_command import DumpGrammarCommand
from strictdoc.commands.manage_autouid_command import ManageAutoUIDCommand
from strictdoc.commands.version_command import VersionCommand
from strictdoc.core.actions.export_action import ExportAction
from strictdoc.core.actions.import_action import ImportAction
from strictdoc.core.project_config import ProjectConfig, ProjectConfigLoader
from strictdoc.helpers.coverage import register_code_coverage_hook
from strictdoc.helpers.exception import (
    ExceptionInfo,
    StrictDocChildProcessException,
)
from strictdoc.helpers.parallelizer import Parallelizer
from strictdoc.helpers.timing import measure_performance
from strictdoc.server.server import run_strictdoc_server


def _main(parallelizer: Parallelizer, parser: SDocArgsParser) -> None:
    register_code_coverage_hook()

    project_config: ProjectConfig

    if parser.is_passthrough_command or parser.is_export_command:
        if parser.is_passthrough_command:
            print(  # noqa: T201
                "warning: passthrough is deprecated, use strictdoc "
                "export --formats sdoc instead."
            )
        export_config: ExportCommandConfig = parser.get_export_config()
        try:
            export_config.validate()
        except CLIValidationError as exception_:
            raise exception_
        project_config = ProjectConfigLoader.load_from_path_or_get_default(
            path_to_config=export_config.get_path_to_config(),
            environment=environment,
        )
        project_config.integrate_export_config(export_config)

        parallelization_value = (
            "Disabled" if export_config.no_parallelization else "Enabled"
        )
        print(  # noqa: T201
            f"Parallelization: {parallelization_value}", flush=True
        )
        export_action = ExportAction(
            project_config=project_config,
            parallelizer=parallelizer,
        )
        export_action.export()

    elif parser.is_server_command:
        server_config = parser.get_server_config()
        try:
            server_config.validate()
        except CLIValidationError as exception_:
            raise exception_
        project_config = ProjectConfigLoader.load_from_path_or_get_default(
            path_to_config=server_config.get_path_to_config(),
            environment=environment,
        )
        run_strictdoc_server(
            server_config=server_config, project_config=project_config
        )

    elif parser.is_import_command_reqif:
        project_config = ProjectConfigLoader.load_from_path_or_get_default(
            path_to_config=os.getcwd(),
            environment=environment,
        )
        import_reqif_config: ImportReqIFCommandConfig = (
            parser.get_import_config_reqif(environment.path_to_strictdoc)
        )
        import_action = ImportAction()
        import_action.do_import(import_reqif_config, project_config)

    elif parser.is_import_command_excel:
        project_config = ProjectConfigLoader.load_from_path_or_get_default(
            path_to_config=os.getcwd(),
            environment=environment,
        )
        import_excel_config: ImportExcelCommandConfig = (
            parser.get_import_config_excel(environment.path_to_strictdoc)
        )
        import_action = ImportAction()
        import_action.do_import(import_excel_config, project_config)

    elif parser.is_manage_autouid_command:
        manage_config: ManageAutoUIDCommandConfig = (
            parser.get_manage_autouid_config()
        )
        try:
            manage_config.validate()
        except CLIValidationError as exception_:
            raise exception_

        project_config = ProjectConfigLoader.load_from_path_or_get_default(
            path_to_config=manage_config.get_path_to_config(),
            environment=environment,
        )
        # FIXME: This must be improved.
        project_config.input_paths = [manage_config.input_path]
        # FIXME: This must be improved.
        project_config.auto_uid_mode = True
        project_config.autouuid_include_sections = (
            manage_config.include_sections
        )

        ManageAutoUIDCommand.execute(
            project_config=project_config, parallelizer=parallelizer
        )

    elif parser.is_dump_grammar_command:
        dump_config: DumpGrammarCommandConfig = parser.get_dump_grammar_config()
        DumpGrammarCommand.execute(dump_config)

    elif parser.is_version_command:
        VersionCommand.execute()

    elif parser.is_about_command:
        AboutCommand.execute()

    elif parser.is_diff_command:
        diff_config: DiffCommandConfig = parser.get_diff_config()
        project_config = ProjectConfigLoader.load_from_path_or_get_default(
            path_to_config=os.getcwd(),
            environment=environment,
        )
        DiffCommand.execute(
            project_config=project_config, diff_config=diff_config
        )

    else:
        raise NotImplementedError


def main() -> None:
    # Ensure that multiprocessing.freeze_support() is called in a frozen
    # application
    # https://github.com/pyinstaller/pyinstaller/issues/7438
    if getattr(sys, "frozen", False):  # pragma: no cover
        multiprocessing.freeze_support()

    # This is crucial for a good performance on macOS. Linux uses 'fork' by default.
    # Changed in version 3.8: On macOS, the spawn start method is now the default.
    # The fork start method should be considered unsafe as it can lead to crashes
    # of the subprocess as macOS system libraries may start threads. See bpo-33725.
    # https://docs.python.org/3/library/multiprocessing.html#contexts-and-start-methods
    # 2024-12-26: StrictDoc has been working with 'fork' just fine, so keep doing it until
    #             anything serious appears against using it.
    if sys.platform != "win32":
        multiprocessing.set_start_method("fork", force=True)
    else:  # pragma: no cover
        pass  # pragma: no cover

    # How to make python 3 print() utf8
    # https://stackoverflow.com/a/3597849/598057
    # sys.stdout.reconfigure(encoding='utf-8') for Python 3.7
    sys.stdout = open(  # pylint: disable=bad-option-value,consider-using-with
        1, "w", encoding="utf-8", closefd=False
    )

    enable_parallelization = "--no-parallelization" not in sys.argv

    # NOTE: The parser can exit before the _main starts when no arguments
    #       or incorrect arguments are provided. In those cases, it is still
    #       important that the parallelizer is correctly shut down.
    try:
        parser = create_sdoc_args_parser()
    except Exception as exception_:
        print(f"error: {str(exception_)}", flush=True)  # noqa: T201
        sys.exit(1)

    if parser.is_debug_mode():
        environment.is_debug_mode = True

    parallelizer = Parallelizer.create(enable_parallelization)

    exception_info: Optional[ExceptionInfo] = None
    try:
        _main(parallelizer, parser)
    except StrictDocChildProcessException as exception_info_:
        exception_info = exception_info_.exception_info
    except Exception as exception_:
        exception_info = ExceptionInfo(exception_)
    finally:
        if exception_info is not None:
            if parser.is_debug_mode():
                print(exception_info.get_stack_trace(), flush=True)  # noqa: T201
            print(exception_info.get_detailed_error_message(), flush=True)  # noqa: T201
            if not parser.is_debug_mode():
                print(  # noqa: T201
                    "Rerun with strictdoc --debug <...> to enable stack trace printing.",
                    flush=True,
                )
            sys.exit(1)


if __name__ == "__main__":
    with measure_performance("Total execution time"):
        main()
else:  # pragma: no cover
    pass



"""
@relation(SDOC-SRS-51, scope=file)
"""

import os
from pathlib import Path
from typing import List, Tuple

from strictdoc.core.project_config import ProjectConfig
from strictdoc.core.traceability_index import TraceabilityIndex
from strictdoc.export.html.generators.document_pdf import (
    DocumentHTML2PDFGenerator,
)
from strictdoc.export.html.html_generator import HTMLGenerator
from strictdoc.export.html.html_templates import HTMLTemplates
from strictdoc.export.html.renderers.link_renderer import LinkRenderer
from strictdoc.export.html.renderers.markup_renderer import MarkupRenderer
from strictdoc.export.html2pdf.pdf_print_driver import PDFPrintDriver
from strictdoc.helpers.exception import StrictDocException
from strictdoc.helpers.git_client import GitClient
from strictdoc.helpers.timing import measure_performance


class HTML2PDFGenerator:
    @staticmethod
    def export_tree(
        project_config: ProjectConfig,
        traceability_index: TraceabilityIndex,
        html_templates: HTMLTemplates,
        output_html2pdf_root: str,
        flat_assets: bool = False,
    ) -> None:
        if not project_config.is_activated_html2pdf():
            raise StrictDocException("HTML2PDF feature is not enabled")

        git_client: GitClient = GitClient()

        path_to_output_pdf_html_dir = os.path.join(output_html2pdf_root, "html")
        path_to_output_pdf_pdf_dir = os.path.join(output_html2pdf_root, "pdf")

        HTMLGenerator.export_assets(
            traceability_index=traceability_index,
            project_config=project_config,
            export_output_html_root=path_to_output_pdf_html_dir,
            flat_assets=flat_assets,
        )

        paths_to_print: List[Tuple[str, str]] = []

        for document_ in traceability_index.document_tree.document_list:
            assert document_.meta is not None

            # Skip generating the included documents, unless the option is provided.
            if not project_config.export_included_documents:
                if document_.document_is_included():
                    continue

            root_path = document_.meta.get_root_path_prefix()

            link_renderer = LinkRenderer(
                root_path=root_path,
                static_path=project_config.dir_for_sdoc_assets,
            )
            markup_renderer = MarkupRenderer.create(
                "RST",
                traceability_index,
                link_renderer,
                html_templates,
                project_config,
                document_,
            )

            with measure_performance("Generating printable HTML document"):
                document_content = DocumentHTML2PDFGenerator.export(
                    project_config,
                    document_,
                    traceability_index,
                    markup_renderer,
                    link_renderer,
                    git_client=git_client,
                    standalone=False,
                    html_templates=html_templates,
                )

            path_to_output_html_doc_dir = os.path.join(
                path_to_output_pdf_html_dir,
                document_.meta.output_document_dir_rel_path.relative_path,
            )
            Path(path_to_output_html_doc_dir).mkdir(parents=True, exist_ok=True)
            Path(path_to_output_pdf_pdf_dir).mkdir(parents=True, exist_ok=True)

            path_to_output_html_doc = os.path.join(
                path_to_output_html_doc_dir,
                document_.meta.document_filename_base + ".html",
            )

            with open(
                path_to_output_html_doc, "w", encoding="utf8"
            ) as output_html_doc_file_:
                output_html_doc_file_.write(document_content)

            path_to_output_pdf_dir = os.path.join(
                path_to_output_pdf_pdf_dir,
                document_.meta.input_doc_dir_rel_path.relative_path,
            )
            Path(path_to_output_pdf_dir).mkdir(parents=True, exist_ok=True)

            path_to_output_pdf = os.path.join(
                path_to_output_pdf_dir,
                document_.meta.document_filename_base + ".pdf",
            )

            paths_to_print.append((path_to_output_html_doc, path_to_output_pdf))

        pdf_print_driver = PDFPrintDriver()
        try:
            pdf_print_driver.get_pdf_from_html(
                project_config,
                paths_to_print,
            )
        except TimeoutError:
            print("error: HTML2PDF: timeout error.")  # noqa: T201



"""
@relation(SDOC-SRS-62, scope=file)
"""

from markupsafe import Markup

from strictdoc.backend.sdoc.models.document import SDocDocument
from strictdoc.core.project_config import ProjectConfig
from strictdoc.core.traceability_index import TraceabilityIndex
from strictdoc.export.html.document_type import DocumentType
from strictdoc.export.html.generators.view_objects.document_screen_view_object import (
    DocumentScreenViewObject,
)
from strictdoc.export.html.html_templates import HTMLTemplates
from strictdoc.export.html.renderers.link_renderer import LinkRenderer
from strictdoc.export.html.renderers.markup_renderer import MarkupRenderer
from strictdoc.helpers.git_client import GitClient


class DocumentTableHTMLGenerator:
    @staticmethod
    def export(
        project_config: ProjectConfig,
        document: SDocDocument,
        traceability_index: TraceabilityIndex,
        markup_renderer: MarkupRenderer,
        link_renderer: LinkRenderer,
        git_client: GitClient,
        html_templates: HTMLTemplates,
    ) -> Markup:
        view_object = DocumentScreenViewObject(
            document_type=DocumentType.TABLE,
            document=document,
            traceability_index=traceability_index,
            project_config=project_config,
            link_renderer=link_renderer,
            markup_renderer=markup_renderer,
            git_client=git_client,
            standalone=False,
        )
        return view_object.render_screen(html_templates.jinja_environment())



"""
@relation(SDOC-SRS-21, scope=file)
"""

from collections import OrderedDict
from typing import Any, Dict, Generator, List, Optional, Tuple, Union

from strictdoc.backend.sdoc.models.model import (
    RESERVED_NON_META_FIELDS,
    RequirementFieldName,
)
from strictdoc.helpers.auto_described import auto_described
from strictdoc.helpers.mid import MID


class RequirementFieldType:
    STRING = "String"
    SINGLE_CHOICE = "SingleChoice"
    MULTIPLE_CHOICE = "MultipleChoice"
    TAG = "Tag"
    REFERENCE = "Reference"


class GrammarReferenceType:
    PARENT_REQ_REFERENCE = "ParentReqReference"
    CHILD_REQ_REFERENCE = "ChildReqReference"
    FILE_REFERENCE = "FileReference"


class ReferenceType:
    PARENT = "Parent"
    CHILD = "Child"
    FILE = "File"

    GRAMMAR_REFERENCE_TYPE_MAP = {
        PARENT: GrammarReferenceType.PARENT_REQ_REFERENCE,
        CHILD: GrammarReferenceType.CHILD_REQ_REFERENCE,
        FILE: GrammarReferenceType.FILE_REFERENCE,
    }


@auto_described
class GrammarElementField:
    def __init__(self) -> None:
        self.title: str = ""
        self.human_title: Optional[str] = None
        self.gef_type: str = ""
        self.required: bool = False
        self.mid: MID = MID.create()

    def get_field_human_name(self) -> str:
        if self.human_title is not None:
            return self.human_title
        return self.title


@auto_described
class GrammarElementFieldString(GrammarElementField):
    def __init__(
        self, parent: Any, title: str, human_title: Optional[str], required: str
    ) -> None:
        super().__init__()
        self.parent: Any = parent
        self.title: str = title
        self.human_title: Optional[str] = human_title
        self.gef_type = RequirementFieldType.STRING
        self.required: bool = required == "True"
        self.mid: MID = MID.create()


@auto_described
class GrammarElementFieldSingleChoice(GrammarElementField):
    def __init__(
        self,
        parent: Any,
        title: str,
        human_title: Optional[str],
        options: List[str],
        required: str,
    ) -> None:
        super().__init__()
        self.parent: Any = parent
        self.title: str = title
        self.human_title: Optional[str] = human_title
        self.gef_type = RequirementFieldType.SINGLE_CHOICE
        self.options: List[str] = options
        self.required: bool = required == "True"
        self.mid: MID = MID.create()


@auto_described
class GrammarElementFieldMultipleChoice(GrammarElementField):
    def __init__(
        self,
        parent: Any,
        title: str,
        human_title: Optional[str],
        options: List[str],
        required: str,
    ) -> None:
        super().__init__()
        self.parent: Any = parent
        self.title: str = title
        self.human_title: Optional[str] = human_title
        self.gef_type = RequirementFieldType.MULTIPLE_CHOICE
        self.options: List[str] = options
        self.required: bool = required == "True"
        self.mid: MID = MID.create()


@auto_described
class GrammarElementFieldTag(GrammarElementField):
    def __init__(
        self, parent: Any, title: str, human_title: Optional[str], required: str
    ) -> None:
        super().__init__()
        self.parent: Any = parent
        self.title: str = title
        self.human_title: Optional[str] = human_title
        self.gef_type = RequirementFieldType.TAG
        self.required: bool = required == "True"
        self.mid: MID = MID.create()


GrammarElementFieldType = Union[
    GrammarElementFieldString,
    GrammarElementFieldSingleChoice,
    GrammarElementFieldMultipleChoice,
    GrammarElementFieldTag,
]


@auto_described
class GrammarElementRelationParent:  # noqa: PLW1641
    def __init__(
        self, parent: Any, relation_type: str, relation_role: Optional[str]
    ) -> None:
        assert relation_type == "Parent"
        self.parent: Any = parent
        self.relation_type: str = relation_type
        self.relation_role: Optional[str] = (
            relation_role
            if relation_role is not None and len(relation_role) > 0
            else None
        )
        self.mid: MID = MID.create()

    def __eq__(self, other: Any) -> bool:
        if not isinstance(other, GrammarElementRelationParent):
            raise AssertionError(self, other)  # pragma: no cover
        return (
            self.mid == other.mid
            and self.relation_type == other.relation_type
            and self.relation_role == other.relation_role
        )


@auto_described
class GrammarElementRelationChild:
    def __init__(
        self, parent: Any, relation_type: str, relation_role: Optional[str]
    ):
        assert relation_type == "Child"
        self.parent: Any = parent
        self.relation_type = relation_type
        self.relation_role: Optional[str] = (
            relation_role
            if relation_role is not None and len(relation_role) > 0
            else None
        )
        self.mid: MID = MID.create()


@auto_described
class GrammarElementRelationFile:
    def __init__(
        self, parent: Any, relation_type: str, relation_role: Optional[str]
    ):
        assert relation_type == "File"
        self.parent: Any = parent
        self.relation_type = relation_type
        self.relation_role: Optional[str] = (
            relation_role
            if relation_role is not None and len(relation_role) > 0
            else None
        )
        self.mid: MID = MID.create()


GrammarElementRelationType = Union[
    GrammarElementRelationParent,
    GrammarElementRelationChild,
    GrammarElementRelationFile,
]


@auto_described()
class GrammarElement:
    def __init__(
        self,
        parent: Any,
        tag: str,
        property_is_composite: str,
        property_prefix: str,
        property_view_style: str,
        fields: List[GrammarElementFieldType],
        relations: List[GrammarElementRelationType],
    ) -> None:
        self.parent: Any = parent
        self.tag: str = tag

        assert property_is_composite in ("", "True", "False")
        self.property_is_composite: Optional[bool] = (
            None
            if property_is_composite == ""
            else (property_is_composite == "True")
        )

        self.property_prefix: Optional[str] = (
            property_prefix if property_prefix not in (None, "") else None
        )

        assert property_view_style in (
            "",
            "Plain",
            "Narrative",
            "Simple",
            "Inline",
            "Table",
            "Zebra",
        )
        self.property_view_style: Optional[str] = (
            property_view_style if property_view_style != "" else None
        )
        self.property_view_style_lower: Optional[str] = (
            property_view_style.lower() if property_view_style != "" else None
        )

        self.fields: List[GrammarElementFieldType] = fields

        self.relations: List[GrammarElementRelationType] = (
            relations if relations is not None and len(relations) > 0 else []
        )

        fields_map: OrderedDict[str, GrammarElementField] = OrderedDict()

        statement_field: Optional[Tuple[str, int]] = None
        description_field: Optional[Tuple[str, int]] = None
        content_field: Optional[Tuple[str, int]] = None
        for field_idx_, field_ in enumerate(fields):
            fields_map[field_.title] = field_
            if field_.title == RequirementFieldName.STATEMENT:
                statement_field = (RequirementFieldName.STATEMENT, field_idx_)
            elif field_.title == "DESCRIPTION":
                description_field = (
                    RequirementFieldName.DESCRIPTION,
                    field_idx_,
                )
            elif field_.title == "CONTENT":
                content_field = (RequirementFieldName.CONTENT, field_idx_)
            else:
                pass
        self.fields_map: Dict[str, GrammarElementField] = fields_map

        self.field_titles: List[str] = list(
            map(lambda field__: field__.title, self.fields)
        )

        self.content_field: Tuple[str, int] = (
            statement_field or description_field or content_field or ("", -1)
        )

        # Use TITLE as a boundary between the single-line and multiline, if
        # TITLE exists. For nodes without a TITLE, use the content field, e.g.,
        # STATEMENT or DESCRIPTION.
        try:
            multiline_field_index = self.get_field_titles().index("TITLE") + 1
        except ValueError:
            multiline_field_index = self.content_field[1]
        self._multiline_field_index: int = multiline_field_index

        self.mid: MID = MID.create()
        self.ng_line_start: Optional[int] = None
        self.ng_col_start: Optional[int] = None

    @staticmethod
    def create_default(tag: str) -> "GrammarElement":
        return GrammarElement(
            parent=None,
            tag=tag,
            property_is_composite="",
            property_prefix="",
            property_view_style="",
            fields=[
                GrammarElementFieldString(
                    parent=None,
                    title="UID",
                    human_title=None,
                    required="False",
                ),
                GrammarElementFieldString(
                    parent=None,
                    title="TITLE",
                    human_title=None,
                    required="False",
                ),
                GrammarElementFieldString(
                    parent=None,
                    title="STATEMENT",
                    human_title=None,
                    required="False",
                ),
            ],
            relations=[],
        )

    @staticmethod
    def create_default_relations(
        parent: "GrammarElement",
    ) -> List[GrammarElementRelationType]:
        return [
            GrammarElementRelationParent(
                parent=parent,
                relation_type="Parent",
                relation_role=None,
            ),
            GrammarElementRelationFile(
                parent=parent,
                relation_type="File",
                relation_role=None,
            ),
        ]

    def is_field_multiline(self, field_name: str) -> bool:
        field_index = self.field_titles.index(field_name)
        return self.is_field_idx_multiline(field_index)

    def is_field_idx_multiline(self, field_idx: int) -> bool:
        # If there is none of TITLE-STATEMENT-DESCRIPTION-CONTENT present, i.e.,
        # multiline_field_index is -1, every field will be treated as multiline.
        return self._multiline_field_index <= field_idx

    def get_view_style(self) -> Optional[str]:
        if self.property_view_style_lower is not None:
            return self.property_view_style_lower
        # For backward compatibility with older versions that didn't have the
        # [[NODE]] syntax and didn't enter the corresponding template migration,
        # keep the TEXT nodes to have a "plain" style unless their type is
        # specified by the grammar.
        if self.tag == "TEXT":
            return "plain"
        return None

    def get_relation_types(self) -> List[str]:
        return list(
            map(lambda relation_: relation_.relation_type, self.relations)
        )

    def get_field_titles(self) -> List[str]:
        return self.field_titles

    def get_tag_lower(self) -> str:
        return self.tag.lower()

    def has_relation_type_role(
        self, relation_type: str, relation_role: Optional[str]
    ) -> bool:
        assert relation_role is None or len(relation_role) > 0
        for relation_ in self.relations:
            if (
                relation_.relation_type == relation_type
                and relation_.relation_role == relation_role
            ):
                return True
        return False

    def enumerate_meta_field_titles(self) -> Generator[str, None, None]:
        for field in self.fields:
            if field.title in (
                RequirementFieldName.TITLE,
                RequirementFieldName.STATEMENT,
            ):
                break
            if field.title in RESERVED_NON_META_FIELDS:
                continue
            yield field.title

    def enumerate_custom_content_field_titles(
        self,
    ) -> Generator[str, None, None]:
        after_title_or_statement = False
        for field in self.fields:
            if field.title in (
                RequirementFieldName.TITLE,
                RequirementFieldName.STATEMENT,
            ):
                after_title_or_statement = True
            if field.title in RESERVED_NON_META_FIELDS:
                continue
            if not after_title_or_statement:
                continue
            yield field.title



"""
@relation(SDOC-SRS-72, scope=file)
"""

import os
from pathlib import Path

from reqif.reqif_bundle import ReqIFZBundle
from reqif.unparser import ReqIFUnparser, ReqIFZUnparser

from strictdoc.backend.reqif.p01_sdoc.sdoc_to_reqif_converter import (
    P01_SDocToReqIFObjectConverter,
)
from strictdoc.backend.reqif.sdoc_reqif_fields import ReqIFProfile
from strictdoc.core.project_config import ProjectConfig
from strictdoc.core.traceability_index import TraceabilityIndex


class ReqIFExport:
    @staticmethod
    def export(
        project_config: ProjectConfig,
        traceability_index: TraceabilityIndex,
        output_reqif_root: str,
        reqifz: bool,
    ) -> None:
        Path(output_reqif_root).mkdir(parents=True, exist_ok=True)

        if project_config.reqif_profile == ReqIFProfile.P01_SDOC:
            reqif_bundle = P01_SDocToReqIFObjectConverter.convert_document_tree(
                document_tree=traceability_index.document_tree,
                multiline_is_xhtml=project_config.reqif_multiline_is_xhtml,
                enable_mid=project_config.reqif_enable_mid,
            )
        else:
            raise NotImplementedError(
                f"Requirements profile does not implement the ReqIF export yet: "
                f"{project_config.reqif_profile}."
            )

        output_file_path: str
        if not reqifz:
            output_file_path = os.path.join(output_reqif_root, "output.reqif")
            reqif_content: str = ReqIFUnparser.unparse(reqif_bundle)
            with open(output_file_path, "w", encoding="utf8") as output_file:
                output_file.write(reqif_content)
        else:
            output_file_path = os.path.join(output_reqif_root, "output.reqifz")
            reqifz_bundle = ReqIFZBundle(
                reqif_bundles={"document_tree.reqif": reqif_bundle},
                attachments={},
            )
            reqifz_content_bytes = ReqIFZUnparser.unparse(reqifz_bundle)
            with open(output_file_path, "wb") as output_file:
                output_file.write(reqifz_content_bytes)



from enum import Enum
from typing import Dict, List, Optional, Union

from markupsafe import Markup

from strictdoc.backend.sdoc.models.document import SDocDocument
from strictdoc.backend.sdoc.models.node import (
    SDocNode,
    SDocNodeField,
)
from strictdoc.backend.sdoc.models.section import SDocSection
from strictdoc.helpers.auto_described import auto_described
from strictdoc.helpers.mid import MID


class ChangeType(str, Enum):
    DOCUMENT = "Document"
    DOCUMENT_MODIFIED = "Document modified"

    SECTION = "Section"
    SECTION_REMOVED = "Section removed"
    SECTION_MODIFIED = "Section modified"
    SECTION_ADDED = "Section added"

    REQUIREMENT = "Requirement"
    REQUIREMENT_REMOVED = "Requirement removed"
    REQUIREMENT_MODIFIED = "Requirement modified"
    REQUIREMENT_ADDED = "Requirement added"


@auto_described
class DocumentChange:
    def __init__(
        self,
        *,
        matched_uid: Optional[str],
        lhs_document: Optional[SDocDocument],
        rhs_document: Optional[SDocDocument],
        uid_modified: bool,
        title_modified: bool,
        lhs_colored_title_diff: Optional[Markup],
        rhs_colored_title_diff: Optional[Markup],
    ):
        assert lhs_document is not None or rhs_document is not None
        if matched_uid is not None:
            assert len(matched_uid) > 0
        self.matched_uid: Optional[str] = matched_uid
        self.uid_modified: bool = uid_modified
        self.title_modified: bool = title_modified
        self.lhs_colored_title_diff: Optional[Markup] = lhs_colored_title_diff
        self.rhs_colored_title_diff: Optional[Markup] = rhs_colored_title_diff

        self.lhs_document: Optional[SDocDocument] = lhs_document
        self.rhs_document: Optional[SDocDocument] = rhs_document

        self.change_type: ChangeType = ChangeType.DOCUMENT_MODIFIED

    def get_colored_title_diff(self, side: str) -> Optional[Markup]:
        assert self.title_modified
        if side == "left":
            return self.lhs_colored_title_diff
        if side == "right":
            return self.rhs_colored_title_diff
        raise AssertionError(f"Must not reach here: {side}")


@auto_described
class SectionChange:
    def __init__(
        self,
        *,
        matched_mid: Optional[MID],
        matched_uid: Optional[str],
        section_token: Optional[str],
        lhs_section: Optional[Union[SDocSection, SDocDocument, SDocNode]],
        rhs_section: Optional[Union[SDocSection, SDocDocument, SDocNode]],
        uid_modified: bool,
        title_modified: bool,
        lhs_colored_title_diff: Optional[Markup],
        rhs_colored_title_diff: Optional[Markup],
    ):
        assert lhs_section is not None or rhs_section is not None
        if matched_uid is not None:
            assert len(matched_uid) > 0
        self.matched_mid: Optional[MID] = matched_mid
        self.matched_uid: Optional[str] = matched_uid
        self.section_token: Optional[str] = section_token
        self.uid_modified: bool = uid_modified
        self.title_modified: bool = title_modified
        self.lhs_colored_title_diff: Optional[Markup] = lhs_colored_title_diff
        self.rhs_colored_title_diff: Optional[Markup] = rhs_colored_title_diff

        self.lhs_section: Optional[
            Union[SDocSection, SDocDocument, SDocNode]
        ] = lhs_section
        self.rhs_section: Optional[
            Union[SDocSection, SDocDocument, SDocNode]
        ] = rhs_section

        if matched_mid is not None or matched_uid is not None:
            change_type = ChangeType.SECTION_MODIFIED
            assert isinstance(lhs_section, SDocSection) or (
                isinstance(lhs_section, SDocNode)
                and lhs_section.node_type == "SECTION"
            ), lhs_section
            assert isinstance(rhs_section, SDocSection) or (
                isinstance(rhs_section, SDocNode)
                and rhs_section.node_type == "SECTION"
            ), rhs_section
        elif lhs_section is not None:
            assert rhs_section is None, rhs_section
            change_type = ChangeType.SECTION_REMOVED
        elif rhs_section is not None:
            assert lhs_section is None, lhs_section
            change_type = ChangeType.SECTION_ADDED
        else:
            raise AssertionError("Must not reach here.")  # pragma: no cover
        self.change_type = change_type

    def is_paired_change(self) -> bool:
        return self.lhs_section is not None and self.rhs_section is not None

    def get_colored_title_diff(self, side: str) -> Optional[Markup]:
        assert self.title_modified
        if side == "left":
            return self.lhs_colored_title_diff
        if side == "right":
            return self.rhs_colored_title_diff
        raise AssertionError(f"Must not reach here: {side}")


@auto_described
class RequirementFieldChange:
    def __init__(
        self,
        *,
        field_name: str,
        lhs_field: Optional[SDocNodeField],
        rhs_field: Optional[SDocNodeField],
        left_diff: Optional[Markup],
        right_diff: Optional[Markup],
    ):
        assert isinstance(field_name, str) and len(field_name) > 0
        assert lhs_field is not None or rhs_field is not None
        assert (left_diff is None and right_diff is None) or (
            left_diff is not None and right_diff is not None
        )

        self.field_name: str = field_name
        self.lhs_field: Optional[SDocNodeField] = lhs_field
        self.rhs_field: Optional[SDocNodeField] = rhs_field
        self.left_diff: Optional[Markup] = left_diff
        self.right_diff: Optional[Markup] = right_diff

    def get_colored_free_text_diff(self, side: str) -> Optional[Markup]:
        if side == "left":
            return self.left_diff
        if side == "right":
            return self.right_diff
        raise AssertionError(f"Must not reach here: {side}")


@auto_described
class RequirementChange:
    def __init__(
        self,
        *,
        requirement_token: Optional[str],
        lhs_requirement: Optional[SDocNode],
        rhs_requirement: Optional[SDocNode],
        field_changes: List[RequirementFieldChange],
    ):
        assert requirement_token is None or len(requirement_token) > 0
        assert lhs_requirement is not None or rhs_requirement is not None

        self.requirement_token: Optional[str] = requirement_token
        self.lhs_requirement: Optional[SDocNode] = lhs_requirement
        self.rhs_requirement: Optional[SDocNode] = rhs_requirement

        self.field_changes: List[RequirementFieldChange] = field_changes

        map_fields_to_changes: Dict[SDocNodeField, RequirementFieldChange] = {}
        for field_change_ in field_changes:
            if field_change_.lhs_field is not None:
                map_fields_to_changes[field_change_.lhs_field] = field_change_
            if field_change_.rhs_field is not None:
                map_fields_to_changes[field_change_.rhs_field] = field_change_
        self.map_fields_to_changes: Dict[
            SDocNodeField, RequirementFieldChange
        ] = map_fields_to_changes

        if requirement_token is not None:
            change_type = ChangeType.REQUIREMENT_MODIFIED
        elif lhs_requirement is not None:
            change_type = ChangeType.REQUIREMENT_REMOVED
        elif rhs_requirement is not None:
            change_type = ChangeType.REQUIREMENT_ADDED
        else:
            raise AssertionError("Must not reach here.")  # pragma: no cover
        self.change_type = change_type

    def is_paired_change(self) -> bool:
        return (
            self.lhs_requirement is not None
            and self.rhs_requirement is not None
        )

    def get_field_change(
        self, requirement_field: SDocNodeField
    ) -> Optional[RequirementFieldChange]:
        assert isinstance(requirement_field, SDocNodeField)
        return self.map_fields_to_changes.get(requirement_field)


ChangeUnionType = Union[DocumentChange, SectionChange, RequirementChange]



from copy import copy
from typing import Any, Dict, List, Optional, Tuple, Type

from strictdoc.core.graph.abstract_bucket import ALL_EDGES, AbstractBucket
from strictdoc.helpers.ordered_set import OrderedSet


class ManyToManySet(AbstractBucket):
    def __init__(self, lhs_type: Type[Any], rhs_type: Type[Any]) -> None:
        self._links: Dict[Any, Dict[Optional[str], OrderedSet[Any]]] = {}
        self._links_reverse: Dict[
            Any, Dict[Optional[str], OrderedSet[Any]]
        ] = {}
        self._lhs_type: type = lhs_type
        self._rhs_type: type = rhs_type

    def has_link(self, *, lhs_node: Any) -> bool:
        assert isinstance(lhs_node, self._lhs_type), lhs_node
        return lhs_node in self._links and len(self._links[lhs_node]) > 0

    def get_count(self, edge: Optional[str] = None) -> int:
        total_count = 0
        for _, lhs_node_links in self._links.items():
            for edge_, lhs_node_edge_links_ in lhs_node_links.items():
                if edge in (ALL_EDGES, edge_):
                    total_count += len(lhs_node_edge_links_)
        return total_count

    def get_link_value(
        self, *, lhs_node: Any, edge: Optional[str] = None
    ) -> Any:
        raise NotImplementedError

    def get_link_value_weak(self, *, lhs_node: Any) -> Any:
        raise NotImplementedError

    def get_link_values(
        self, *, lhs_node: Any, edge: Optional[str] = ALL_EDGES
    ) -> OrderedSet[Any]:
        assert isinstance(lhs_node, self._lhs_type), lhs_node
        if edge == ALL_EDGES:
            all_values: OrderedSet[Any] = OrderedSet()
            for _, edge_links_ in self._links.setdefault(lhs_node, {}).items():
                for edge_link_ in edge_links_:
                    all_values.add(edge_link_)
            return all_values
        else:
            return self._links.setdefault(lhs_node, {}).setdefault(
                edge, OrderedSet()
            )

    def get_link_values_with_edges(
        self, *, lhs_node: Any, edge: Optional[str] = ALL_EDGES
    ) -> List[Tuple[Any, Optional[str]]]:
        assert isinstance(lhs_node, self._lhs_type), lhs_node
        all_values: List[Tuple[Any, Optional[str]]] = []
        for edge_, edge_links_ in self._links.setdefault(lhs_node, {}).items():
            if edge in (ALL_EDGES, edge_):
                for edge_link_ in edge_links_:
                    all_values.append((edge_link_, edge_))
        return all_values

    def get_link_values_reverse(
        self, *, rhs_node: Any, edge: Optional[str] = ALL_EDGES
    ) -> OrderedSet[Any]:
        assert isinstance(rhs_node, self._rhs_type), rhs_node
        if edge == ALL_EDGES:
            all_values: OrderedSet[Any] = OrderedSet()
            for _, edge_links_ in self._links_reverse.setdefault(
                rhs_node, {}
            ).items():
                for edge_link_ in edge_links_:
                    all_values.add(edge_link_)
            return all_values
        return self._links_reverse.setdefault(rhs_node, {}).setdefault(
            edge, OrderedSet()
        )

    def create_link(
        self, *, lhs_node: Any, rhs_node: Any, edge: Optional[str] = None
    ) -> None:
        assert edge != ALL_EDGES
        if not isinstance(lhs_node, self._lhs_type):
            raise TypeError(
                f"LHS type mismatch: {type(lhs_node)} is not of {self._lhs_type}"
            )
        if not isinstance(rhs_node, self._rhs_type):
            raise TypeError(
                f"RHS type mismatch: {type(rhs_node)} is not of {self._rhs_type}"
            )
        assert lhs_node != rhs_node, (lhs_node, rhs_node)

        lhs_node_links = self._links.setdefault(lhs_node, {})
        lhs_node_links_for_edge = lhs_node_links.setdefault(edge, OrderedSet())

        assert rhs_node not in lhs_node_links_for_edge
        lhs_node_links_for_edge.add(rhs_node)

        rhs_node_links = self._links_reverse.setdefault(rhs_node, {})
        rhs_node_links_for_edge = rhs_node_links.setdefault(edge, OrderedSet())
        assert lhs_node not in rhs_node_links_for_edge
        rhs_node_links_for_edge.add(lhs_node)

    def delete_link(
        self,
        *,
        lhs_node: Any,
        rhs_node: Any,
        edge: Optional[str] = ALL_EDGES,
    ) -> None:
        assert isinstance(lhs_node, self._lhs_type), lhs_node
        assert isinstance(rhs_node, self._rhs_type), rhs_node

        assert lhs_node in self._links
        assert rhs_node in self._links_reverse

        lhs_node_links = self._links[lhs_node]
        if edge == ALL_EDGES:
            for _, edge_links_ in lhs_node_links.items():
                edge_links_.discard(rhs_node)
        else:
            assert edge in lhs_node_links
            lhs_node_links[edge].remove(rhs_node)

        rhs_node_links = self._links_reverse[rhs_node]
        if edge == ALL_EDGES:
            for _, edge_links_ in rhs_node_links.items():
                edge_links_.discard(lhs_node)
        else:
            assert edge in rhs_node_links
            rhs_node_links[edge].remove(lhs_node)

    def delete_link_weak(
        self, *, lhs_node: Any, rhs_node: Any, edge: Optional[str] = ALL_EDGES
    ) -> None:
        assert isinstance(lhs_node, self._lhs_type), lhs_node
        assert isinstance(rhs_node, self._rhs_type), rhs_node

        if lhs_node in self._links:
            lhs_node_links = self._links.get(lhs_node, {})
            if edge == ALL_EDGES:
                for _, edge_links_ in lhs_node_links.items():
                    edge_links_.discard(rhs_node)
            else:
                lhs_node_links_for_edge = lhs_node_links.get(edge, OrderedSet())
                lhs_node_links_for_edge.discard(rhs_node)

        if rhs_node in self._links_reverse:
            rhs_node_links = self._links_reverse.get(rhs_node, {})
            if edge == ALL_EDGES:
                for _, edge_links_ in rhs_node_links.items():
                    edge_links_.discard(lhs_node)
            else:
                rhs_node_links_for_edge = rhs_node_links.get(edge, OrderedSet())
                rhs_node_links_for_edge.discard(lhs_node)

    def delete_all_links(
        self,
        *,
        lhs_node: Any,
    ) -> None:
        assert isinstance(lhs_node, self._lhs_type), lhs_node

        assert lhs_node in self._links

        del self._links[lhs_node]

        for existing_rhs_node_ in copy(self._links_reverse):
            existing_rhs_node_links = self._links_reverse[existing_rhs_node_]
            for _, edge_links_ in existing_rhs_node_links.items():
                edge_links_.discard(lhs_node)



"""
@relation(SDOC-SRS-142, scope=file)
"""

from functools import partial
from typing import Any, Dict, List, Optional, TypedDict, Union

from textx import get_location, metamodel_from_str

from strictdoc.backend.sdoc.error_handling import StrictDocSemanticError
from strictdoc.backend.sdoc_source_code.grammar import SOURCE_FILE_GRAMMAR
from strictdoc.backend.sdoc_source_code.models.function_range_marker import (
    FunctionRangeMarker,
)
from strictdoc.backend.sdoc_source_code.models.line_marker import LineMarker
from strictdoc.backend.sdoc_source_code.models.range_marker import (
    RangeMarker,
)
from strictdoc.backend.sdoc_source_code.models.requirement_marker import Req
from strictdoc.backend.sdoc_source_code.models.source_file_info import (
    SourceFileTraceabilityInfo,
)
from strictdoc.backend.sdoc_source_code.parse_context import ParseContext
from strictdoc.helpers.file_stats import SourceFileStats
from strictdoc.helpers.list import find_duplicates
from strictdoc.helpers.textx import drop_textx_meta


class TextXLocation(TypedDict):
    line: int
    col: int
    filename: str


def req_processor(req: Req) -> None:
    assert isinstance(req, Req), (
        f"Expected req to be Req, got: {req}, {type(req)}"
    )
    location = get_location(req)
    assert location
    req.ng_source_line = location["line"]
    req.ng_source_column = location["col"]


def source_file_traceability_info_processor(
    source_file_traceability_info: SourceFileTraceabilityInfo,
    parse_context: ParseContext,
) -> None:
    if len(parse_context.marker_stack) > 0:
        raise create_unmatch_range_error(
            parse_context.marker_stack,
        )
    source_file_traceability_info.markers = parse_context.markers
    source_file_traceability_info.file_stats = parse_context.file_stats


def create_begin_end_range_reqs_mismatch_error(
    location: Dict[str, Any],
    lhs_marker_reqs: List[str],
    rhs_marker_reqs: List[str],
) -> StrictDocSemanticError:
    lhs_marker_reqs_str = ", ".join(lhs_marker_reqs)
    rhs_marker_reqs_str = ", ".join(rhs_marker_reqs)

    return StrictDocSemanticError(
        "STRICTDOC RANGE: BEGIN and END requirements mismatch",
        (
            "STRICT RANGE marker should START and END "
            "with the same requirement(s): "
            f"'{lhs_marker_reqs_str}' != '{rhs_marker_reqs_str}'."
        ),
        # @sdoc[nosdoc]  # noqa: ERA001
        """
# [REQ-001]
Content...
# [/REQ-001]
        """.lstrip(),
        # @sdoc[/nosdoc]  # noqa: ERA001
        line=location["line"],
        col=location["col"],
        filename=location["filename"],
    )


def create_end_without_begin_error(
    location: Dict[str, Any],
) -> StrictDocSemanticError:
    return StrictDocSemanticError(
        "STRICTDOC RANGE: END marker without preceding BEGIN marker",
        (
            "STRICT RANGE shall be opened with "
            "START marker and ended with END marker."
        ),
        # @sdoc[nosdoc]  # noqa: ERA001
        """
# [REQ-001]
Content...
# [/REQ-001]
        """.lstrip(),
        # @sdoc[/nosdoc]  # noqa: ERA001
        line=location["line"],
        col=location["col"],
        filename=location["filename"],
    )


def create_unmatch_range_error(
    unmatched_ranges: List[RangeMarker],
) -> StrictDocSemanticError:
    assert isinstance(unmatched_ranges, list)
    assert len(unmatched_ranges) > 0
    range_locations: List[TextXLocation] = []
    for unmatched_range in unmatched_ranges:
        location = get_location(unmatched_range)
        range_locations.append(location)
    first_location = range_locations[0]
    hint: Optional[str] = None
    if len(unmatched_ranges) > 1:
        range_lines = list(map(lambda loc: loc["line"], range_locations[1:]))
        hint = f"The @sdoc keywords are also unmatched on lines: {range_lines}."

    return StrictDocSemanticError(
        "Unmatched @sdoc keyword found in source file.",
        hint=hint,
        # @sdoc[nosdoc]
        example=(
            "Each @sdoc keyword must be matched with a closing keyword. "
            "Example:\n"
            "@sdoc[REQ-001]\n"
            "...\n"
            "@sdoc[/REQ-001]"
        ),
        # @sdoc[/nosdoc]
        line=first_location["line"],
        col=first_location["col"],
        filename=first_location["filename"],
    )


def validate_marker_uids(
    marker: Union[FunctionRangeMarker, LineMarker, RangeMarker],
    parse_context: ParseContext,
) -> None:
    possible_duplicates = find_duplicates(marker.reqs)
    if len(possible_duplicates) > 0:
        location = get_location(marker)

        raise ValueError(
            "@relation marker contains duplicate node UIDs: "
            f"{possible_duplicates}. Location: {parse_context.filename}:{location['line']}."
        )


def range_marker_processor(
    marker: RangeMarker, parse_context: ParseContext
) -> None:
    validate_marker_uids(marker, parse_context)

    location = get_location(marker)
    line = location["line"]

    current_top_marker: RangeMarker
    if marker.ng_is_nodoc:
        if marker.is_begin():
            parse_context.marker_stack.append(marker)
        elif marker.is_end():
            try:
                current_top_marker = parse_context.marker_stack.pop()
                if (
                    not current_top_marker.ng_is_nodoc
                    or current_top_marker.is_end()
                ):
                    raise create_begin_end_range_reqs_mismatch_error(
                        location, current_top_marker.reqs, marker.reqs
                    )
            except IndexError:
                raise create_end_without_begin_error(location) from None
        return

    if (
        len(parse_context.marker_stack) > 0
        and parse_context.marker_stack[-1].ng_is_nodoc
    ):
        # This marker is within a "nosdoc" block, so we ignore it.
        return

    parse_context.markers.append(marker)
    marker.ng_source_line_begin = line

    if marker.is_begin():
        marker.ng_range_line_begin = line
        parse_context.marker_stack.append(marker)
        for req in marker.reqs:
            markers = parse_context.map_reqs_to_markers.setdefault(req, [])
            markers.append(marker)

    elif marker.is_end():
        try:
            current_top_marker = parse_context.marker_stack.pop()
            if marker.reqs != current_top_marker.reqs:
                raise create_begin_end_range_reqs_mismatch_error(
                    location, current_top_marker.reqs, marker.reqs
                )
            current_top_marker.ng_range_line_end = line

            marker.ng_range_line_begin = current_top_marker.ng_range_line_begin
            marker.ng_range_line_end = line

        except IndexError:
            raise create_end_without_begin_error(location) from None
    else:
        raise NotImplementedError


def line_marker_processor(
    line_marker: LineMarker, parse_context: ParseContext
) -> None:
    validate_marker_uids(line_marker, parse_context)

    location = get_location(line_marker)
    line = location["line"]

    if (
        len(parse_context.marker_stack) > 0
        and parse_context.marker_stack[-1].ng_is_nodoc
    ):
        # This marker is within a "nosdoc" block, so we ignore it.
        return

    parse_context.markers.append(line_marker)
    line_marker.ng_source_line_begin = line
    line_marker.ng_range_line_begin = line
    line_marker.ng_range_line_end = line

    for req in line_marker.reqs:
        markers = parse_context.map_reqs_to_markers.setdefault(req, [])
        markers.append(line_marker)


def function_range_marker_processor(
    function_range_marker: FunctionRangeMarker, parse_context: ParseContext
) -> None:
    validate_marker_uids(function_range_marker, parse_context)

    if (
        len(parse_context.marker_stack) > 0
        and parse_context.marker_stack[-1].ng_is_nodoc
    ):
        # This marker is within a "nosdoc" block, so we ignore it.
        return

    parse_context.markers.append(function_range_marker)
    function_range_marker.ng_source_line_begin = 1
    function_range_marker.ng_range_line_begin = 1
    function_range_marker.ng_range_line_end = (
        parse_context.file_stats.lines_total
    )

    for req in function_range_marker.reqs:
        markers = parse_context.map_reqs_to_markers.setdefault(req, [])
        markers.append(function_range_marker)


class SourceFileTraceabilityReader:
    SOURCE_FILE_MODELS = [
        FunctionRangeMarker,
        LineMarker,
        Req,
        SourceFileTraceabilityInfo,
        RangeMarker,
    ]

    def __init__(self) -> None:
        self.meta_model = metamodel_from_str(
            SOURCE_FILE_GRAMMAR,
            classes=SourceFileTraceabilityReader.SOURCE_FILE_MODELS,
            use_regexp_group=True,
        )

    def read(
        self, input_string: str, file_path: Optional[str] = None
    ) -> SourceFileTraceabilityInfo:
        # TODO: This might be possible to handle directly in the textx grammar.
        # AttributeError: 'str' object has no attribute '_tx_parser'.
        file_size = len(input_string)
        if file_size == 0:
            return SourceFileTraceabilityInfo([])

        file_stats = SourceFileStats.create(input_string)
        parse_context = ParseContext(file_path, file_stats)

        parse_source_traceability_processor = partial(
            source_file_traceability_info_processor, parse_context=parse_context
        )
        parse_req_processor = partial(req_processor)
        parse_range_marker_processor = partial(
            range_marker_processor, parse_context=parse_context
        )
        parse_line_marker_processor = partial(
            line_marker_processor, parse_context=parse_context
        )
        parse_function_range_marker_processor = partial(
            function_range_marker_processor, parse_context=parse_context
        )

        obj_processors = {
            "FunctionRangeMarker": parse_function_range_marker_processor,
            "LineMarker": parse_line_marker_processor,
            "RangeMarker": parse_range_marker_processor,
            "Req": parse_req_processor,
            "SourceFileTraceabilityInfo": parse_source_traceability_processor,
        }

        self.meta_model.register_obj_processors(obj_processors)

        source_file_traceability_info: SourceFileTraceabilityInfo = (
            self.meta_model.model_from_str(input_string, file_name=file_path)
        )
        source_file_traceability_info.ng_map_reqs_to_markers = (
            parse_context.map_reqs_to_markers
        )

        # HACK:
        # ProcessPoolExecutor doesn't work because of non-picklable parts
        # of textx. The offending fields are stripped down because they
        # are not used anyway.
        drop_textx_meta(source_file_traceability_info)
        return source_file_traceability_info

    def read_from_file(self, file_path: str) -> SourceFileTraceabilityInfo:
        with open(file_path, encoding="utf-8") as file:
            sdoc_content = file.read()
            sdoc = self.read(sdoc_content, file_path=file_path)
            return sdoc



from collections import deque
from typing import Callable, Deque, List, Set, Tuple

from strictdoc.backend.sdoc.errors.document_tree_error import DocumentTreeError

BEFORE = 1
AFTER = 2


class TreeCycleDetector:
    def __init__(self) -> None:
        self.checked: Set[str] = set()

    def check_node(
        self, node: str, links_function: Callable[[str], List[str]]
    ) -> None:
        if node in self.checked:
            return
        stack: Deque[Tuple[str, int]] = deque()
        stack.append((node, BEFORE))
        visited: Set[str] = set()
        while stack:
            current_node, token = stack[-1]
            if token == BEFORE:
                visited.add(current_node)
                stack[-1] = (current_node, AFTER)
                node_links: List[str] = links_function(current_node)
                for node_link in reversed(node_links):
                    if node_link in visited:
                        cycled_nodes = []
                        for uid, token in stack:
                            if token == BEFORE:
                                continue
                            cycled_nodes.append(uid)
                        raise DocumentTreeError.cycle_error(
                            node_link, cycled_nodes
                        )
                    if node_link not in self.checked:
                        stack.append((node_link, BEFORE))
            elif token == AFTER:
                visited.remove(current_node)
                self.checked.add(current_node)
                stack.pop()
        self.checked.add(node)


class SingleShotTreeCycleDetector:
    def check_node(
        self,
        new_uid: str,
        node: str,
        links_function: Callable[[str], List[str]],
    ) -> None:
        """
        Detect if a new node creates a cycle in the traceability index.

        FIXME: Both cycle detector classes are very similar. Find a way to merge
               them. The added value of this 'single shot' detector is that it
               checks for cycles possibly added by a new node.
        """

        checked: Set[str] = set()

        stack: Deque[Tuple[str, int]] = deque()
        stack.append((node, BEFORE))
        visited = {new_uid}
        while stack:
            current_node, token = stack[-1]
            if token == BEFORE:
                visited.add(current_node)
                stack[-1] = (current_node, AFTER)
                node_links = links_function(current_node)
                for node_link in reversed(node_links):
                    if node_link in visited:
                        cycled_nodes = [new_uid]
                        for uid, token in stack:
                            if token == BEFORE:
                                continue
                            cycled_nodes.append(uid)
                        raise DocumentTreeError.cycle_error(
                            node_link, cycled_nodes
                        )
                    if node_link not in checked:
                        stack.append((node_link, BEFORE))
            elif token == AFTER:
                visited.remove(current_node)
                checked.add(current_node)
                stack.pop()



import os
from pathlib import Path
from typing import List

from strictdoc.core.file_tree import File, FileFinder
from strictdoc.core.project_config import ProjectConfig
from strictdoc.core.source_tree import SourceFile, SourceTree


class SourceFilesFinder:
    @staticmethod
    def find_source_files(
        project_config: ProjectConfig,
    ) -> SourceTree:
        map_file_to_source = {}
        found_source_files: List[SourceFile] = []

        # TODO: Unify this on the FileTree class level.
        doctree_root_abs_path = (
            project_config.source_root_path
            if project_config.source_root_path is not None
            else os.getcwd()
        )
        doctree_root_abs_path = (
            os.path.dirname(doctree_root_abs_path)
            if os.path.isfile(doctree_root_abs_path)
            else doctree_root_abs_path
        )

        assert isinstance(project_config.output_dir, str)

        exclude_paths = project_config.exclude_source_paths + ["**.DS_Store"]

        file_tree = FileFinder.find_files_with_extensions(
            root_path=doctree_root_abs_path,
            ignored_dirs=[project_config.output_dir],
            extensions=None,
            include_paths=project_config.include_source_paths,
            exclude_paths=exclude_paths,
        )
        root_level = doctree_root_abs_path.count(os.sep)

        file: File
        for _, file, _ in file_tree.iterate():
            in_doctree_source_file_rel_path = os.path.relpath(
                file.full_path, doctree_root_abs_path
            )
            last_folder_in_path: str = os.path.relpath(
                file.folder_path, doctree_root_abs_path
            )
            output_dir_full_path: str = os.path.join(
                project_config.export_output_html_root,
                "_source_files",
                last_folder_in_path,
            )
            Path(output_dir_full_path).mkdir(parents=True, exist_ok=True)

            output_file_name = f"{file.file_name}.html"
            output_file_full_path = os.path.join(
                output_dir_full_path, output_file_name
            )

            level = file.folder_path.count(os.sep) - root_level

            source_file = SourceFile(
                level,
                file.full_path,
                in_doctree_source_file_rel_path,
                output_dir_full_path,
                output_file_full_path,
            )
            found_source_files.append(source_file)
            map_file_to_source[file] = source_file

        source_tree = SourceTree(
            file_tree, found_source_files, map_file_to_source
        )
        return source_tree



import os
import urllib
from copy import deepcopy
from datetime import datetime
from typing import Optional

from fastapi import APIRouter
from starlette.responses import HTMLResponse, Response

from strictdoc import __version__
from strictdoc.core.project_config import ProjectConfig
from strictdoc.export.html.document_type import DocumentType
from strictdoc.export.html.generators.view_objects.diff_screen_results_view_object import (
    DiffScreenResultsViewObject,
)
from strictdoc.export.html.generators.view_objects.diff_screen_view_object import (
    DiffScreenViewObject,
)
from strictdoc.export.html.html_templates import HTMLTemplates
from strictdoc.export.html.renderers.link_renderer import LinkRenderer
from strictdoc.git.change_generator import ChangeContainer, ChangeGenerator
from strictdoc.git.git_client import GitClient
from strictdoc.server.routers.main_router import (
    HTTP_STATUS_BAD_REQUEST,
    HTTP_STATUS_PRECONDITION_FAILED,
)


def create_other_router(project_config: ProjectConfig) -> APIRouter:
    router = APIRouter()

    html_templates = HTMLTemplates.create(
        project_config=project_config,
        enable_caching=False,
        strictdoc_last_update=datetime.today(),
    )

    @router.get("/diff")
    def get_git_diff(
        left_revision: Optional[str] = None,
        right_revision: Optional[str] = None,
        tab: Optional[str] = None,
    ) -> HTMLResponse:
        if not project_config.is_activated_diff():
            return Response(
                content="The DIFF feature is not activated in the project config.",
                status_code=HTTP_STATUS_PRECONDITION_FAILED,
            )
        if tab is not None:
            if tab not in ("diff", "changelog"):
                return Response(
                    content="The tab= parameter must be either 'diff' or 'changelog'.",
                    status_code=HTTP_STATUS_BAD_REQUEST,
                )
        else:
            tab = "diff"

        error_message: Optional[str] = None

        if (
            left_revision is not None
            and len(left_revision) > 0
            and right_revision is not None
            and len(right_revision) > 0
        ):
            git_client = GitClient(".")
            try:
                if left_revision != "HEAD+":
                    git_client.check_revision(left_revision)
                else:
                    raise LookupError(
                        "Left revision argument 'HEAD+' is not supported. "
                        "'HEAD+' can only be used as a right revision argument."
                    )

                if right_revision != "HEAD+":
                    git_client.check_revision(right_revision)
            except LookupError as exception_:
                error_message = exception_.args[0]
        elif (left_revision is not None and len(left_revision) > 0) or (
            right_revision is not None and len(right_revision) > 0
        ):
            error_message = "Valid Git revisions must be provided."
        else:
            # In the case when both revisions are empty, we load the starting
            # diff page.
            pass

        view_object = DiffScreenViewObject(
            project_config=project_config,
            results=False,
            left_revision=left_revision,
            right_revision=right_revision,
            error_message=error_message,
            tab=tab,
        )
        output = view_object.render_screen(html_templates.jinja_environment())
        status_code = 200 if error_message is None else 422
        return HTMLResponse(content=output, status_code=status_code)

    @router.get("/diff_result")
    def get_git_diff_result(
        left_revision: Optional[str] = None,
        right_revision: Optional[str] = None,
        tab: Optional[str] = None,
    ) -> HTMLResponse:
        if not project_config.is_activated_diff():
            return Response(
                content="The DIFF feature is not activated in the project config.",
                status_code=HTTP_STATUS_PRECONDITION_FAILED,
            )
        if tab is not None and tab not in ("diff", "changelog"):
            return Response(
                content="The tab= parameter must be either 'diff' or 'changelog'.",
                status_code=HTTP_STATUS_PRECONDITION_FAILED,
            )
        elif tab is None:
            tab = "diff"

        left_revision_resolved = None
        right_revision_resolved = None

        results = False
        error_message: Optional[str] = None

        if (
            left_revision is not None
            and len(left_revision) > 0
            and right_revision is not None
            and len(right_revision) > 0
        ):
            git_client = GitClient(".")
            try:
                if left_revision != "HEAD+":
                    left_revision_resolved = git_client.check_revision(
                        left_revision
                    )
                else:
                    raise LookupError(
                        "Left revision argument 'HEAD+' is not supported. "
                        "'HEAD+' can only be used as a right revision argument."
                    )

                if right_revision == "HEAD+":
                    right_revision_resolved = "HEAD+"
                else:
                    right_revision_resolved = git_client.check_revision(
                        right_revision
                    )

                results = True
            except LookupError as exception_:
                error_message = exception_.args[0]
        elif (left_revision is not None and len(left_revision) > 0) or (
            right_revision is not None and len(right_revision) > 0
        ):
            error_message = "Valid Git revisions must be provided."
        else:
            # In the case when both revisions are empty, we load the starting
            # diff page.
            pass

        path_to_template = (
            "screens/git/frame_changelog_result.jinja"
            if tab == "changelog"
            else "screens/git/frame_diff_result.jinja"
        )
        template = html_templates.jinja_environment().get_template(
            path_to_template
        )

        link_renderer = LinkRenderer(
            root_path="", static_path=project_config.dir_for_sdoc_assets
        )

        left_revision_urlencoded = (
            urllib.parse.quote(left_revision)
            if left_revision is not None
            else ""
        )
        right_revision_urlencoded = (
            urllib.parse.quote(right_revision)
            if right_revision is not None
            else ""
        )

        if not results:
            output = template.render(
                project_config=project_config,
                document_type=DocumentType.DOCUMENT.value,
                link_document_type=DocumentType.DOCUMENT.value,
                standalone=False,
                strictdoc_version=__version__,
                link_renderer=link_renderer,
                results=False,
                left_revision=left_revision,
                left_revision_urlencoded=left_revision_urlencoded,
                right_revision=right_revision,
                right_revision_urlencoded=right_revision_urlencoded,
                error_message=error_message,
            )
            status_code = 200 if error_message is None else 422
            return HTMLResponse(content=output, status_code=status_code)

        assert left_revision_resolved is not None
        assert right_revision_resolved is not None

        git_client_lhs = GitClient.create_repo_from_local_copy(
            left_revision_resolved, project_config
        )

        project_config_copy_lhs: ProjectConfig = deepcopy(project_config)
        assert project_config_copy_lhs.input_paths is not None
        project_config_copy_rhs: ProjectConfig = deepcopy(project_config)
        assert project_config_copy_rhs.input_paths is not None

        export_input_rel_path = os.path.relpath(
            project_config_copy_lhs.input_paths[0], os.getcwd()
        )
        export_input_abs_path = os.path.join(
            git_client_lhs.path_to_git_root, export_input_rel_path
        )
        project_config_copy_lhs.input_paths = [export_input_abs_path]

        git_client_rhs = GitClient.create_repo_from_local_copy(
            right_revision_resolved, project_config
        )

        export_input_rel_path = os.path.relpath(
            project_config_copy_rhs.input_paths[0], os.getcwd()
        )
        export_input_abs_path = os.path.join(
            git_client_rhs.path_to_git_root, export_input_rel_path
        )
        project_config_copy_rhs.input_paths = [export_input_abs_path]

        change_container: ChangeContainer = ChangeGenerator.generate(
            lhs_project_config=project_config_copy_lhs,
            rhs_project_config=project_config_copy_rhs,
        )

        assert change_container.traceability_index_lhs.document_tree is not None
        assert change_container.traceability_index_rhs.document_tree is not None
        view_object = DiffScreenResultsViewObject(
            project_config=project_config,
            change_container=change_container,
            document_tree_lhs=change_container.traceability_index_lhs.document_tree,
            document_tree_rhs=change_container.traceability_index_rhs.document_tree,
            documents_iterator_lhs=change_container.documents_iterator_lhs,
            documents_iterator_rhs=change_container.documents_iterator_rhs,
            left_revision=left_revision,
            right_revision=right_revision,
            lhs_stats=change_container.lhs_stats,
            rhs_stats=change_container.rhs_stats,
            change_stats=change_container.change_stats,
            traceability_index_lhs=change_container.traceability_index_lhs,
            traceability_index_rhs=change_container.traceability_index_rhs,
            tab=tab,
        )
        output = template.render(view_object=view_object)

        return HTMLResponse(
            content=output,
            status_code=200,
        )

    return router



import pickle
from typing import Any, Optional


def pickle_dump(obj: Any) -> bytes:
    return pickle.dumps(obj, 0)


def pickle_load(content: bytes) -> Optional[Any]:
    try:
        return pickle.loads(content)
    except AttributeError:
        # Known issue: when SDoc schema, e.g., core classes, changes,
        # unpicking results in Exceptions. So far, only the AttributeError
        # was observed.
        return None



"""
View Object class that bridges between project metrics data and the Project
Statistics screen Jinja templates.

@relation(SDOC-SRS-97, scope=file)
"""

from dataclasses import dataclass
from typing import List, Union

from markupsafe import Markup

from strictdoc import __version__
from strictdoc.core.document_tree import DocumentTree
from strictdoc.core.document_tree_iterator import DocumentTreeIterator
from strictdoc.core.project_config import ProjectConfig
from strictdoc.core.statistics.metric import Metric, MetricSection
from strictdoc.core.traceability_index import TraceabilityIndex
from strictdoc.export.html.html_templates import JinjaEnvironment
from strictdoc.export.html.renderers.link_renderer import LinkRenderer
from strictdoc.helpers.cast import assert_cast


@dataclass
class ProjectStatisticsViewObject:
    def __init__(
        self,
        *,
        traceability_index: TraceabilityIndex,
        project_config: ProjectConfig,
        link_renderer: LinkRenderer,
        metrics: List[Union[Metric, MetricSection]],
    ) -> None:
        self.traceability_index: TraceabilityIndex = traceability_index
        self.project_config: ProjectConfig = project_config
        self.link_renderer: LinkRenderer = link_renderer
        self.metrics: List[Union[Metric, MetricSection]] = metrics

        self.document_tree_iterator: DocumentTreeIterator = (
            DocumentTreeIterator(
                assert_cast(traceability_index.document_tree, DocumentTree)
            )
        )
        self.standalone: bool = False
        self.is_running_on_server: bool = project_config.is_running_on_server
        self.strictdoc_version = __version__

    def get_document_level(self) -> int:
        return 0

    def render_screen(self, jinja_environment: JinjaEnvironment) -> Markup:
        return jinja_environment.render_template_as_markup(
            "screens/project_statistics/index.jinja", view_object=self
        )

    def render_static_url(self, url: str) -> Markup:
        return Markup(self.link_renderer.render_static_url(url))

    def render_url(self, url: str) -> Markup:
        return Markup(self.link_renderer.render_url(url))



import os
import sys
from pathlib import Path
from typing import Optional

from strictdoc.backend.excel.export.excel_generator import ExcelGenerator
from strictdoc.backend.reqif.reqif_export import ReqIFExport
from strictdoc.backend.sdoc.errors.document_tree_error import DocumentTreeError
from strictdoc.backend.sdoc.models.document import SDocDocument
from strictdoc.backend.sdoc.writer import SDWriter
from strictdoc.core.project_config import ProjectConfig
from strictdoc.core.traceability_index import TraceabilityIndex
from strictdoc.core.traceability_index_builder import TraceabilityIndexBuilder
from strictdoc.export.doxygen.doxygen_generator import DoxygenGenerator
from strictdoc.export.html.document_type import DocumentType
from strictdoc.export.html.html_generator import HTMLGenerator
from strictdoc.export.html.html_templates import HTMLTemplates
from strictdoc.export.html2pdf.html2pdf_generator import HTML2PDFGenerator
from strictdoc.export.json.json_generator import JSONGenerator
from strictdoc.export.rst.document_rst_generator import DocumentRSTGenerator
from strictdoc.export.spdx.spdx_generator import SPDXGenerator
from strictdoc.helpers.parallelizer import NullParallelizer, Parallelizer
from strictdoc.helpers.timing import timing_decorator


class ExportAction:
    def __init__(
        self,
        project_config: ProjectConfig,
        parallelizer: Parallelizer,
    ):
        assert parallelizer
        self.project_config: ProjectConfig = project_config
        self.parallelizer = parallelizer
        self.traceability_index: TraceabilityIndex = self.build_index()

    @timing_decorator("Parse SDoc project tree")
    def build_index(self) -> TraceabilityIndex:
        try:
            traceability_index: TraceabilityIndex = (
                TraceabilityIndexBuilder.create(
                    project_config=self.project_config,
                    parallelizer=self.parallelizer,
                )
            )
        except DocumentTreeError as exc:
            print(exc.to_print_message())  # noqa: T201
            sys.exit(1)
        self.traceability_index = traceability_index
        return traceability_index

    @timing_decorator("Export SDoc")
    def export(self) -> None:
        assert self.traceability_index is not None, (
            "The index must be built at this point."
        )
        assert self.project_config.export_formats is not None, (
            "The export_formats must not be None."
        )

        if (
            "html" in self.project_config.export_formats
            or "html-standalone" in self.project_config.export_formats
            or "html2pdf" in self.project_config.export_formats
        ):
            is_small_project = self.traceability_index.is_small_project()

            html_templates = HTMLTemplates.create(
                project_config=self.project_config,
                enable_caching=not is_small_project,
                strictdoc_last_update=self.traceability_index.strictdoc_last_update,
            )

            # The bundle document is generated only when the option is provided.
            traceability_index_copy: Optional[TraceabilityIndex] = None
            bundle_document: Optional[SDocDocument] = None
            if self.project_config.generate_bundle_document:
                traceability_index_copy, bundle_document = (
                    self.traceability_index.clone_to_bundle_document(
                        self.project_config
                    )
                )

            if (
                "html" in self.project_config.export_formats
                or "html-standalone" in self.project_config.export_formats
            ):
                html_generator = HTMLGenerator(
                    self.project_config, html_templates
                )
                html_generator.export_complete_tree(
                    traceability_index=self.traceability_index,
                    parallelizer=self.parallelizer,
                )
                if self.project_config.generate_bundle_document:
                    assert bundle_document is not None, (
                        "bundle_document must not be None."
                    )
                    assert traceability_index_copy is not None, (
                        "traceability_index_copy must not be None."
                    )
                    html_generator.export_single_document(
                        document=bundle_document,
                        traceability_index=traceability_index_copy,
                        specific_documents=(DocumentType.DOCUMENT,),
                    )

            if "html2pdf" in self.project_config.export_formats:
                output_html2pdf_root = os.path.join(
                    self.project_config.output_dir, "html2pdf"
                )
                Path(output_html2pdf_root).mkdir(parents=True, exist_ok=True)
                HTML2PDFGenerator.export_tree(
                    self.project_config,
                    self.traceability_index,
                    html_templates,
                    output_html2pdf_root,
                )

                if self.project_config.generate_bundle_document:
                    assert traceability_index_copy is not None, (
                        "traceability_index_copy must not be None."
                    )
                    HTML2PDFGenerator.export_tree(
                        self.project_config,
                        traceability_index_copy,
                        html_templates,
                        output_html2pdf_root,
                        flat_assets=True,
                    )

        if "rst" in self.project_config.export_formats:
            output_rst_root = os.path.join(
                self.project_config.output_dir, "rst"
            )
            Path(output_rst_root).mkdir(parents=True, exist_ok=True)
            DocumentRSTGenerator.export_tree(
                self.traceability_index, output_rst_root
            )

        if "excel" in self.project_config.export_formats:
            output_excel_root = f"{self.project_config.output_dir}/excel"
            ExcelGenerator.export_tree(
                self.traceability_index,
                output_excel_root,
                project_config=self.project_config,
            )

        if "reqif-sdoc" in self.project_config.export_formats:
            output_reqif_root = f"{self.project_config.output_dir}/reqif"
            ReqIFExport.export(
                project_config=self.project_config,
                traceability_index=self.traceability_index,
                output_reqif_root=output_reqif_root,
                reqifz=False,
            )

        if "reqifz-sdoc" in self.project_config.export_formats:
            output_reqif_root = f"{self.project_config.output_dir}/reqif"
            ReqIFExport.export(
                project_config=self.project_config,
                traceability_index=self.traceability_index,
                output_reqif_root=output_reqif_root,
                reqifz=True,
            )

        if "sdoc" in self.project_config.export_formats:
            self.export_sdoc()

        if "doxygen" in self.project_config.export_formats:
            output_doxygen_root = os.path.join(
                self.project_config.output_dir, "doxygen"
            )
            doxygen_generator = DoxygenGenerator(
                project_config=self.project_config
            )
            doxygen_generator.export(
                traceability_index=self.traceability_index,
                path_to_output_dir=output_doxygen_root,
            )

        if "spdx" in self.project_config.export_formats:
            output_dot_root = os.path.join(
                self.project_config.output_dir, "spdx"
            )
            Path(output_dot_root).mkdir(parents=True, exist_ok=True)
            SPDXGenerator().export_tree(
                self.project_config, self.traceability_index, output_dot_root
            )

        if "json" in self.project_config.export_formats:
            output_json_root = os.path.join(
                self.project_config.output_dir, "json"
            )
            Path(output_json_root).mkdir(parents=True, exist_ok=True)
            JSONGenerator().export_tree(
                self.traceability_index, self.project_config, output_json_root
            )

    def export_sdoc(self) -> None:
        assert self.project_config.input_paths
        try:
            traceability_index: TraceabilityIndex = (
                TraceabilityIndexBuilder.create(
                    project_config=self.project_config,
                    parallelizer=NullParallelizer(),
                )
            )
        except DocumentTreeError as exc:
            print(exc.to_print_message())  # noqa: T201
            sys.exit(1)
        else:
            assert traceability_index.document_tree

        writer = SDWriter(self.project_config)

        output_base_dir = (
            self.project_config.output_dir
            if self.project_config.output_dir is not None
            else os.path.join(os.getcwd(), "output")
        )
        output_dir = os.path.join(output_base_dir, "sdoc")
        for document in traceability_index.document_tree.document_list:
            assert document.meta
            assert document.meta.document_filename_base
            assert document.meta.input_doc_dir_rel_path
            output, fragments_dict = writer.write_with_fragments(
                document,
            )

            path_to_output_file_dir: str = os.path.join(
                output_dir, document.meta.input_doc_dir_rel_path.relative_path
            )
            Path(path_to_output_file_dir).mkdir(parents=True, exist_ok=True)
            path_to_output_file = os.path.join(
                path_to_output_file_dir, document.meta.document_filename_base
            )
            path_to_output_file += ".sdoc"
            with open(path_to_output_file, "w", encoding="utf8") as file:
                file.write(output)

            for fragment_path_, fragment_content_ in fragments_dict.items():
                path_to_output_fragment = os.path.join(
                    path_to_output_file_dir, fragment_path_
                )
                with open(
                    path_to_output_fragment, "w", encoding="utf8"
                ) as file_:
                    file_.write(fragment_content_)



from datetime import datetime
from typing import List, Union

from markupsafe import Markup

from strictdoc.backend.sdoc.models.node import SDocNode
from strictdoc.backend.sdoc.models.section import SDocSection
from strictdoc.backend.sdoc_source_code.models.source_file_info import (
    SourceFileTraceabilityInfo,
)
from strictdoc.core.document_iterator import DocumentCachingIterator
from strictdoc.core.project_config import ProjectConfig
from strictdoc.core.statistics.metric import Metric, MetricSection
from strictdoc.core.traceability_index import TraceabilityIndex
from strictdoc.export.html.generators.view_objects.project_statistics_view_object import (
    ProjectStatisticsViewObject,
)
from strictdoc.export.html.generators.view_objects.project_tree_stats import (
    DocumentTreeStats,
)
from strictdoc.export.html.html_templates import HTMLTemplates
from strictdoc.export.html.renderers.link_renderer import LinkRenderer
from strictdoc.helpers.cast import assert_cast
from strictdoc.helpers.git_client import GitClient


class ProgressStatisticsGenerator:
    @staticmethod
    def export(
        project_config: ProjectConfig,
        traceability_index: TraceabilityIndex,
        link_renderer: LinkRenderer,
        html_templates: HTMLTemplates,
    ) -> Markup:
        """
        Export default project statistics to an HTML page.

        @relation(SDOC-SRS-97, scope=function)
        """

        git_client = GitClient()

        document_tree_stats: DocumentTreeStats = DocumentTreeStats()

        document_tree_stats.total_documents = len(
            traceability_index.document_tree.document_list
        )
        document_tree_stats.git_commit_hash = git_client.get_commit_hash()

        for document in traceability_index.document_tree.document_list:
            document_iterator = DocumentCachingIterator(document)
            for node, _ in document_iterator.all_content(print_fragments=False):
                if isinstance(node, SDocSection) or (
                    isinstance(node, SDocNode) and node.node_type == "SECTION"
                ):
                    document_tree_stats.total_sections += 1
                    if not node.has_any_text_nodes():
                        document_tree_stats.sections_without_text_nodes += 1

                elif isinstance(node, SDocNode):
                    if (
                        node.is_normative_node()
                        and node.node_type == "REQUIREMENT"
                    ):
                        requirement: SDocNode = assert_cast(node, SDocNode)
                        document_tree_stats.total_requirements += 1

                        if requirement.reserved_uid is None:
                            document_tree_stats.requirements_no_uid += 1

                        if requirement.reserved_status != "Backlog":
                            if document.config.root:
                                if (
                                    len(
                                        traceability_index.get_children_requirements(
                                            requirement
                                        )
                                    )
                                    == 0
                                ):
                                    document_tree_stats.requirements_root_no_links += 1
                            else:
                                if (
                                    len(
                                        traceability_index.get_parent_requirements(
                                            requirement
                                        )
                                    )
                                    == 0
                                ):
                                    document_tree_stats.requirements_no_links += 1

                        # RATIONALE.
                        if (
                            requirement.ordered_fields_lookup.get("RATIONALE")
                            is None
                        ):
                            document_tree_stats.requirements_no_rationale += 1

                        # STATUS.
                        if requirement.reserved_status is None:
                            document_tree_stats.requirements_status_breakdown[
                                None
                            ] += 1
                        else:
                            document_tree_stats.requirements_status_breakdown[
                                requirement.reserved_status
                            ] += 1

                    for requirement_field_ in node.enumerate_fields():
                        field_value = requirement_field_.get_text_value()
                        if "TBD" in field_value:
                            document_tree_stats.total_tbd += 1
                        if "TBC" in field_value:
                            document_tree_stats.total_tbc += 1

        document_tree_stats.sort_requirements_status_breakdown()

        if traceability_index.document_tree.source_tree is not None:
            for (
                source_file_
            ) in traceability_index.document_tree.source_tree.source_files:
                document_tree_stats.total_source_files += 1
                source_file_info_: SourceFileTraceabilityInfo = traceability_index.get_file_traceability_index().get_coverage_info(
                    source_file_.in_doctree_source_file_rel_path_posix
                )
                if source_file_info_.get_coverage() == 100:
                    document_tree_stats.total_source_files_complete_coverage += 1
                elif source_file_info_.get_coverage() == 0:
                    document_tree_stats.total_source_files_no_coverage += 1
                else:
                    document_tree_stats.total_source_files_partial_coverage += 1

        #
        # Create all metrics.
        #

        metrics: List[Union[Metric, MetricSection]] = []

        section = MetricSection(name="General information", metrics=[])
        metrics.append(section)

        section.metrics.append(
            Metric(name="Project name", value=project_config.project_title)
        )
        section.metrics.append(
            Metric(
                name="Statistics generation date",
                value=datetime.today().strftime("%Y-%m-%d %H:%M:%S"),
            )
        )
        section.metrics.append(
            Metric(
                name="Last modification of project data",
                value=traceability_index.strictdoc_last_update.strftime(
                    "%Y-%m-%d %H:%M:%S"
                ),
            )
        )
        section.metrics.append(
            Metric(
                name="Git commit/release",
                value=document_tree_stats.git_commit_hash,
            )
        )
        section.metrics.append(
            Metric(
                name="Total documents",
                value=str(document_tree_stats.total_documents),
            )
        )

        section = MetricSection(name="Sections", metrics=[])
        metrics.append(section)

        section.metrics.append(
            Metric(
                name="Total sections",
                value=str(document_tree_stats.total_sections),
                link="search?q=node.is_section()",
            )
        )
        section.metrics.append(
            Metric(
                name="Sections without any text",
                value=str(document_tree_stats.sections_without_text_nodes),
                link="search?q=(node.is_section() and not node.contains_any_text)",
            )
        )

        section = MetricSection(name="Requirements", metrics=[])
        metrics.append(section)

        section.metrics.append(
            Metric(
                name="Total requirements",
                value=str(document_tree_stats.total_requirements),
                link="search?q=node.is_requirement()",
            )
        )
        section.metrics.append(
            Metric(
                name="Requirements with no UID",
                value=str(document_tree_stats.requirements_no_uid),
                link='search?q=(node.is_requirement() and node["UID"] == None)',
            )
        )
        section.metrics.append(
            Metric(
                name="Root-level requirements not connected to by any requirement",
                value=str(document_tree_stats.requirements_root_no_links),
                link='search?q=(node.is_requirement() and node.is_root and node["STATUS"] != "Backlog" and not node.has_child_requirements)',
            )
        )
        section.metrics.append(
            Metric(
                name="Non-root-level requirements not connected to any parent requirement",
                value=str(document_tree_stats.requirements_no_links),
                link='search?q=(node.is_requirement() and not node.is_root and node["STATUS"] != "Backlog" and not node.has_parent_requirements)',
            )
        )
        section.metrics.append(
            Metric(
                name="Requirements with no RATIONALE",
                value=str(document_tree_stats.requirements_no_rationale),
                link='search?q=(node.is_requirement() and node["RATIONALE"] == None)',
            )
        )

        #
        # If there are requirement statuses, collect status metrics.
        #
        statuses = document_tree_stats.requirements_status_breakdown.keys()
        if any(key is not None for key in statuses):
            section = MetricSection(
                name="Requirements status breakdown", metrics=[]
            )
            metrics.append(section)

            for (
                status_,
                status_count_,
            ) in document_tree_stats.requirements_status_breakdown.items():
                status_query_value = (
                    f'"{status_}"' if status_ is not None else "None"
                )
                section.metrics.append(
                    Metric(
                        name=f"Requirements with status {status_}",
                        value=str(status_count_),
                        link='search?q=(node.is_requirement() and node["STATUS"] == '
                        + status_query_value
                        + ")",
                    )
                )

        #
        # Source files metrics.
        #

        section = MetricSection(name="Source files", metrics=[])
        metrics.append(section)

        section.metrics.append(
            Metric(
                name="Total source files",
                value=str(document_tree_stats.total_source_files),
                link="search?q=node.is_source_file()",
            )
        )
        section.metrics.append(
            Metric(
                name="Total source files with complete requirements coverage",
                value=str(
                    document_tree_stats.total_source_files_complete_coverage
                ),
                link="search?q=node.is_source_file_with_complete_coverage()",
            )
        )
        section.metrics.append(
            Metric(
                name="Total source files with partial requirements coverage",
                value=str(
                    document_tree_stats.total_source_files_partial_coverage
                ),
                link="search?q=node.is_source_file_with_partial_coverage()",
            )
        )
        section.metrics.append(
            Metric(
                name="Total source files with no requirements coverage",
                value=str(document_tree_stats.total_source_files_no_coverage),
                link="search?q=node.is_source_file_with_no_coverage()",
            )
        )

        #
        # TBD/TBC metrics.
        #

        section = MetricSection(name="TBD/TBC", metrics=[])
        metrics.append(section)

        section.metrics.append(
            Metric(
                name="Total TBD",
                value=str(document_tree_stats.total_tbd),
                link='search?q=node.contains("TBD")',
            )
        )
        section.metrics.append(
            Metric(
                name="Total TBC",
                value=str(document_tree_stats.total_tbc),
                link='search?q=node.contains("TBC")',
            )
        )

        #
        # Return the final view object.
        #
        view_object = ProjectStatisticsViewObject(
            traceability_index=traceability_index,
            project_config=project_config,
            link_renderer=link_renderer,
            metrics=metrics,
        )
        return view_object.render_screen(html_templates.jinja_environment())



from enum import IntEnum


class GraphLinkType(IntEnum):
    MID_TO_NODE = 1
    UID_TO_NODE = 2
    NODE_TO_PARENT_NODES = 3
    NODE_TO_CHILD_NODES = 4
    NODE_TO_INCOMING_LINKS = 5
    DOCUMENT_TO_TAGS = 8
